from typing import Optional

USER_ROOT = 1
USER_SYSTEM_ID = -1
USER_SYSTEM_NAME = "Polyaxon"


def is_system_user(user_id: Optional[int] = None) -> bool:
    if user_id:
        return user_id == USER_SYSTEM_ID

    return False



from haupt.common import conf
from haupt.common.options.registry import core

conf.subscribe(core.Logging)
conf.subscribe(core.Debug)
conf.subscribe(core.Protocol)
conf.subscribe(core.CeleryBrokerBackend)
conf.subscribe(core.CeleryBrokerUrl)
conf.subscribe(core.SecretInternalToken)
conf.subscribe(core.HealthCheckWorkerTimeout)
conf.subscribe(core.SchedulerEnabled)
conf.subscribe(core.UiAdminEnabled)
conf.subscribe(core.UiSingleUrl)
conf.subscribe(core.UiDefaultStreamsUrl)
conf.subscribe(core.UiAssetsVersion)
conf.subscribe(core.UiBaseUrl)
conf.subscribe(core.UiOffline)
conf.subscribe(core.UiEnabled)
conf.subscribe(core.UiInSandbox)



from typing import Dict, Optional, Union

from clipped.utils.bools import to_bool
from clipped.utils.json import orjson_loads
from rest_framework import status

from django.core.handlers.asgi import ASGIRequest
from django.db import transaction
from django.http import FileResponse, HttpResponse
from django.urls import path

from haupt.common.endpoints.files import FilePathResponse
from haupt.common.endpoints.validation import validate_methods
from haupt.streams.connections.fs import AppFS
from haupt.streams.controllers.notebooks import render_notebook
from haupt.streams.controllers.uploads import handle_posted_data
from haupt.streams.endpoints.base import UJSONResponse
from haupt.streams.endpoints.utils import redirect_file
from polyaxon._fs.async_manager import (
    check_is_file,
    delete_file_or_dir,
    download_dir,
    download_file,
    list_files,
)


async def handle_upload(
    request: ASGIRequest, run_uuid: str, is_file: bool
) -> HttpResponse:
    content_file = request.FILES.get("upload_file")
    if not content_file:
        return HttpResponse(
            content="A file is required to upload a file",
            status=status.HTTP_400_BAD_REQUEST,
        )
    content_json = request.POST.get("json")
    content_json = orjson_loads(content_json) if content_json else {}
    overwrite = content_json.get("overwrite", True)
    untar = content_json.get("untar", True)
    path = content_json.get("path", "")
    connection = content_json.get("connection")
    try:
        archived_path = await handle_posted_data(
            fs=await AppFS.get_fs(connection=connection),
            store_path=AppFS.get_fs_root_path(connection=connection),
            content_file=content_file,
            root_path=run_uuid,
            path=path,
            upload=True,
            is_file=is_file,
            overwrite=overwrite,
            untar=untar,
        )
    except Exception as e:
        return HttpResponse(
            content="Run's artifacts upload was unsuccessful, "
            "an error was raised while uploading the data %s." % e,
            status=status.HTTP_400_BAD_REQUEST,
        )
    if not archived_path:
        return HttpResponse(
            content="Artifact not found and not uploaded: filepath={}".format(
                archived_path
            ),
            status=status.HTTP_404_NOT_FOUND,
        )
    return HttpResponse(status=status.HTTP_200_OK)


def clean_path(filepath: str):
    return filepath.strip("/")


@transaction.non_atomic_requests
async def handle_artifact(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    project: str,
    run_uuid: str,
    methods: Optional[Dict] = None,
) -> HttpResponse:
    validate_methods(request, methods)
    if request.method == "GET":
        return await download_artifact(request, run_uuid=run_uuid)
    if request.method == "DELETE":
        return await delete_artifact(request, run_uuid=run_uuid)
    if request.method == "POST":
        return await upload_artifact(request, run_uuid=run_uuid)


@transaction.non_atomic_requests
async def ro_artifact(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    project: str,
    run_uuid: str,
    path: str,
    methods: Optional[Dict] = None,
) -> HttpResponse:
    validate_methods(request, methods)
    return await download_artifact(request, run_uuid, path, True)


async def download_artifact(
    request: ASGIRequest,
    run_uuid: str,
    path: Optional[str] = None,
    stream: Optional[bool] = None,
) -> Union[HttpResponse, FileResponse]:
    filepath = request.GET.get("path", path or "")
    stream = to_bool(request.GET.get("stream", stream), handle_none=True)
    force = to_bool(request.GET.get("force"), handle_none=True)
    render = to_bool(request.GET.get("render"), handle_none=True)
    connection = request.GET.get("connection")
    if not filepath:
        return HttpResponse(
            content="A `path` query param is required to stream a file content",
            status=status.HTTP_400_BAD_REQUEST,
        )
    if render and not filepath.endswith(".ipynb"):
        return HttpResponse(
            content="Artifact with 'filepath={}' does not have a valid extension.".format(
                filepath
            ),
            status=status.HTTP_400_BAD_REQUEST,
        )
    subpath = "{}/{}".format(run_uuid, clean_path(filepath)).rstrip("/")
    archived_path = await download_file(
        fs=await AppFS.get_fs(connection=connection),
        store_path=AppFS.get_fs_root_path(connection=connection),
        subpath=subpath,
        check_cache=not force,
    )
    if not archived_path:
        return HttpResponse(
            content="Artifact not found: filepath={}".format(subpath),
            status=status.HTTP_404_NOT_FOUND,
        )
    if stream:
        if render:
            archived_path = await render_notebook(
                archived_path=archived_path, check_cache=not force
            )
        return FilePathResponse(filepath=archived_path)
    return await redirect_file(archived_path)


async def upload_artifact(request: ASGIRequest, run_uuid: str) -> HttpResponse:
    return await handle_upload(request=request, run_uuid=run_uuid, is_file=True)


async def delete_artifact(request: ASGIRequest, run_uuid: str) -> HttpResponse:
    filepath = request.GET.get("path", "")
    if not filepath:
        return HttpResponse(
            content="A `path` query param is required to delete a file",
            status=status.HTTP_400_BAD_REQUEST,
        )
    subpath = "{}/{}".format(run_uuid, clean_path(filepath)).rstrip("/")
    connection = request.GET.get("connection")
    is_deleted = await delete_file_or_dir(
        fs=await AppFS.get_fs(connection=connection),
        store_path=AppFS.get_fs_root_path(connection=connection),
        subpath=subpath,
        is_file=True,
    )
    if not is_deleted:
        return HttpResponse(
            content="Artifact could not be deleted: filepath={}".format(subpath),
            status=status.HTTP_400_BAD_REQUEST,
        )
    return HttpResponse(
        status=status.HTTP_204_NO_CONTENT,
    )


@transaction.non_atomic_requests
async def handle_artifacts(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    project: str,
    run_uuid: str,
    methods: Optional[Dict] = None,
) -> HttpResponse:
    validate_methods(request, methods)
    if request.method == "GET":
        return await download_artifacts(request, run_uuid=run_uuid)
    if request.method == "DELETE":
        return await delete_artifacts(request, run_uuid=run_uuid)
    if request.method == "POST":
        return await upload_artifacts(request, run_uuid=run_uuid)


async def download_artifacts(request: ASGIRequest, run_uuid: str) -> HttpResponse:
    check_path = to_bool(request.GET.get("check_path"), handle_none=True)
    if not check_path:
        # Backwards compatibility
        check_path = to_bool(request.GET.get("check_file"), handle_none=True)
    path = request.GET.get("path", "")
    connection = request.GET.get("connection")
    subpath = "{}/{}".format(run_uuid, clean_path(path)).rstrip("/")
    fs = await AppFS.get_fs(connection=connection)
    store_path = AppFS.get_fs_root_path(connection=connection)
    if check_path:
        is_file = await check_is_file(fs=fs, store_path=store_path, subpath=subpath)
        if is_file:
            return await download_artifact(request, run_uuid=run_uuid)
    archived_path = await download_dir(
        fs=fs, store_path=store_path, subpath=subpath, to_tar=True
    )
    if not archived_path:
        return HttpResponse(
            content="Artifact not found: filepath={}".format(archived_path),
            status=status.HTTP_404_NOT_FOUND,
        )
    return await redirect_file(archived_path)


async def upload_artifacts(request: ASGIRequest, run_uuid: str) -> HttpResponse:
    return await handle_upload(request=request, run_uuid=run_uuid, is_file=False)


async def delete_artifacts(request: ASGIRequest, run_uuid: str) -> HttpResponse:
    path = request.GET.get("path", "")
    connection = request.GET.get("connection")
    subpath = "{}/{}".format(run_uuid, clean_path(path)).rstrip("/")
    is_deleted = await delete_file_or_dir(
        fs=await AppFS.get_fs(connection=connection),
        store_path=AppFS.get_fs_root_path(connection=connection),
        subpath=subpath,
        is_file=False,
    )
    if not is_deleted:
        return HttpResponse(
            content="Artifacts could not be deleted: filepath={}".format(subpath),
            status=status.HTTP_400_BAD_REQUEST,
        )
    return HttpResponse(
        status=status.HTTP_204_NO_CONTENT,
    )


@transaction.non_atomic_requests
async def tree_artifacts(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    project: str,
    run_uuid: str,
    methods: Optional[Dict] = None,
) -> UJSONResponse:
    validate_methods(request, methods)
    filepath = request.GET.get("path", "")
    connection = request.GET.get("connection")
    ls = await list_files(
        fs=await AppFS.get_fs(connection=connection),
        store_path=AppFS.get_fs_root_path(connection=connection),
        subpath=run_uuid,
        filepath=clean_path(filepath),
        force=True,
    )
    return UJSONResponse(ls)


URLS_RUNS_ARTIFACT = (
    "<str:namespace>/<str:owner>/<str:project>/runs/<str:run_uuid>/artifact"
)
URLS_RUNS_EMBEDDED_ARTIFACT = (
    "<str:namespace>/<str:owner>/<str:project>/runs/<str:run_uuid>/embedded_artifact"
)
URLS_RUNS_RO_ARTIFACT = "<str:namespace>/<str:owner>/<str:project>/runs/<str:run_uuid>/ro_artifact/<path:path>"
URLS_RUNS_ARTIFACTS = (
    "<str:namespace>/<str:owner>/<str:project>/runs/<str:run_uuid>/artifacts"
)
URLS_RUNS_ARTIFACTS_TREE = (
    "<str:namespace>/<str:owner>/<str:project>/runs/<str:run_uuid>/artifacts/tree"
)

# fmt: off
artifacts_routes = [
    path(
        URLS_RUNS_ARTIFACT,
        handle_artifact,
        name="artifact",
        kwargs=dict(methods=["GET", "DELETE", "POST"]),
    ),
    path(
        URLS_RUNS_EMBEDDED_ARTIFACT,
        handle_artifact,
        name="download_embedded_artifact",
        kwargs=dict(methods=["GET"]),
    ),
    path(
        URLS_RUNS_RO_ARTIFACT,
        ro_artifact,
        name="read_only_artifact",
        kwargs=dict(methods=["GET"]),
    ),
    path(
        URLS_RUNS_ARTIFACTS,
        handle_artifacts,
        name="download_artifacts",
        kwargs=dict(methods=["GET", "DELETE", "POST"]),
    ),
    path(
        URLS_RUNS_ARTIFACTS_TREE,
        tree_artifacts,
        name="list_artifacts",
        kwargs=dict(methods=["GET"]),
    ),
]



import tempfile

from django.conf import settings
from django.core.cache import cache
from django.test import TestCase


class PolyaxonBaseTest(TestCase):
    COLLECT_TASKS = False

    def setUp(self):
        # Flush cache
        cache.clear()
        # Mock celery default sent task
        self.mock_send_task()
        super().setUp()
        self.worker_send = {}
        settings.ARTIFACTS_ROOT = tempfile.mkdtemp()
        settings.ARCHIVES_ROOT = tempfile.mkdtemp()

    def mock_send_task(self):
        from celery import current_app

        def send_task(name, args=(), kwargs=None, **opts):
            kwargs = kwargs or {}
            if name in current_app.tasks:
                task = current_app.tasks[name]
                return task.apply_async(args, kwargs, **opts)
            elif self.worker_send:
                self.worker_send[name] = {"args": args, "kwargs": kwargs, "opts": opts}

        current_app.send_task = send_task


class PolyaxonBaseTestSerializer(PolyaxonBaseTest):
    query = None
    serializer_class = None
    model_class = None
    factory_class = None
    expected_keys = {}
    num_objects = 2

    def test_serialize_one(self):
        raise NotImplementedError

    def create_one(self):
        raise NotImplementedError

    def create_multiple(self):
        for i in range(self.num_objects):
            self.create_one()

    def test_serialize_many(self):
        self.create_multiple()
        data = self.serializer_class(self.query.all(), many=True).data
        assert len(data) == self.num_objects
        for d in data:
            assert set(d.keys()) == self.expected_keys



from clipped.compact.pydantic import ValidationError as PydanticValidationError
from clipped.utils.json import orjson_loads
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from django.conf import settings
from django.db.models import Q

from haupt.common.authentication.base import is_normal_user
from haupt.db.defs import Models
from haupt.db.managers.live_state import archive_run as base_archive_run
from haupt.db.managers.live_state import restore_run as base_restore_run
from haupt.db.managers.runs import add_run_contributors, base_approve_run
from haupt.db.managers.statuses import (
    new_run_skipped_status,
    new_run_status,
    new_run_stopping_status,
)
from polyaxon.exceptions import PolyaxonException
from polyaxon.schemas import V1StatusCondition


def clone_run(view, request, *args, **kwargs):
    view.run = view.get_object()
    run = view.pre_validate(view.run)
    content = None
    if "content" in request.data:
        content = request.data["content"]
    if content and not isinstance(content, dict):
        try:
            content = orjson_loads(content)
        except Exception as e:
            raise ValidationError("Cloning was not successful, error: {}".format(e))
    try:
        new_obj = view.clone(
            obj=run,
            content=content,
            name=request.data.get("name"),
            description=request.data.get("description"),
            tags=request.data.get("tags"),
            meta_info=request.data.get("meta_info"),
        )
    except (PydanticValidationError, PolyaxonException, ValueError) as e:
        raise ValidationError("Cloning was not successful, error: {}".format(e))

    add_run_contributors(new_obj, users=[request.user])
    view.audit(request, *args, **kwargs)
    serializer = view.get_serializer(new_obj)
    return Response(status=status.HTTP_201_CREATED, data=serializer.data)


def create_status(view, serializer):
    serializer.is_valid()
    validated_data = serializer.validated_data
    if not validated_data:
        return
    condition = None
    if validated_data.get("condition"):
        condition = V1StatusCondition.get_condition(**validated_data.get("condition"))
    if not condition:
        return
    if settings.HAS_ORG_MANAGEMENT and is_normal_user(view.request.user):
        status_meta_info = {
            "user": {
                "username": view.request.user.username,
                "email": view.request.user.email,
            },
        }
        condition.meta_info = status_meta_info
    new_run_status(
        run=view.run, condition=condition, force=validated_data.get("force", False)
    )


def stop_run(view, request, *args, **kwargs):
    status_meta_info = None
    if settings.HAS_ORG_MANAGEMENT and is_normal_user(request.user):
        status_meta_info = {
            "user": {
                "username": view.request.user.username,
                "email": view.request.user.email,
            },
        }
    if new_run_stopping_status(
        run=view.run,
        message="User requested to stop the run.",
        meta_info=status_meta_info,
    ):
        add_run_contributors(view.run, users=[request.user])
        view.audit(request, *args, **kwargs)
    return Response(status=status.HTTP_200_OK, data={})


def skip_run(view, request, *args, **kwargs):
    status_meta_info = None
    if settings.HAS_ORG_MANAGEMENT and is_normal_user(request.user):
        status_meta_info = {
            "user": {
                "username": view.request.user.username,
                "email": view.request.user.email,
            },
        }
    if new_run_skipped_status(
        run=view.run,
        message="User requested to skip the run.",
        meta_info=status_meta_info,
    ):
        add_run_contributors(view.run, users=[request.user])
        view.audit(request, *args, **kwargs)
    return Response(status=status.HTTP_200_OK, data={})


def approve_run(view, request, *args, **kwargs):
    pending = view.run.pending
    if pending:
        base_approve_run(view.run)
        add_run_contributors(view.run, users=[request.user])
        view.audit(request, *args, **kwargs)
    return Response(status=status.HTTP_200_OK, data={})


def transfer_run(view, request, *args, **kwargs):
    project_name = request.data.get("project")
    if project_name == view.project_name:
        return Response(status=status.HTTP_200_OK, data={})
    if not project_name:
        raise ValidationError("The destination project was not provided.")
    try:
        filters = {"name": project_name}
        if settings.HAS_ORG_MANAGEMENT:
            filters["owner__name"] = view.owner_name
        dest_project = Models.Project.objects.only("id").get(**filters)
    except Models.Project.DoesNotExist:
        raise ValidationError(
            "The destination project `{}` does not exist.".format(project_name)
        )

    view.run.project_id = dest_project.id
    view.run.save(update_fields=["project_id", "updated_at"])
    if view.run.has_pipeline:
        Models.Run.all.filter(Q(pipeline=view.run) | Q(controller=view.run)).update(
            project_id=dest_project.id
        )
    add_run_contributors(view.run, users=[request.user])
    view.audit(request, *args, **kwargs)
    return Response(status=status.HTTP_200_OK, data={})


def invalidate_run(view, request, *args, **kwargs):
    view.run.state = None
    view.run.save(update_fields=["state"])
    add_run_contributors(view.run, users=[request.user])
    view.audit(request, *args, **kwargs)
    return Response(status=status.HTTP_200_OK, data={})


def archive_run(view, request, *args, **kwargs):
    view.run = view.get_object()
    view.audit(request, *args, **kwargs)
    add_run_contributors(view.run, users=[request.user])
    base_archive_run(view.run)
    return Response(status=status.HTTP_200_OK, data={})


def restore_run(view, request, *args, **kwargs):
    view.run = view.get_object()
    view.audit(request, *args, **kwargs)
    add_run_contributors(view.run, users=[request.user])
    base_restore_run(view.run)
    return Response(status=status.HTTP_200_OK, data={})



from haupt import settings
from haupt.proxies.schemas.base import get_config

CORS_OPTIONS = r"""
    if ($request_method = 'OPTIONS') {{
        add_header 'Access-Control-Allow-Origin' "$http_origin" always;
        add_header 'Access-Control-Allow-Credentials' 'true' always;
        add_header 'Access-Control-Allow-Headers' 'X-POLYAXON-SERVICE,x-csrftoken,X-CSRF-Token,Authorization,Accept,Origin,DNT,X-Mx-ReqToken,X-CustomHeader,Keep-Alive,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Content-Range,Range,Sec-Fetch-Mode,User-Agent' always;
        add_header 'Access-Control-Allow-Methods' 'GET,POST,OPTIONS,PUT,DELETE,PATCH' always;
        add_header 'Access-Control-Max-Age' 1728000 always;
        add_header 'Content-Length' 0;
        add_header 'Content-Type' 'text/plain; charset=utf-8';
        return 204;
    }}

    add_header 'Access-Control-Allow-Origin' "$http_origin" always;
    add_header 'Access-Control-Allow-Credentials' 'true' always;
    add_header 'Access-Control-Allow-Methods' 'GET,POST,OPTIONS,PUT,DELETE,PATCH' always;
    add_header 'Access-Control-Allow-Headers' 'X-POLYAXON-SERVICE,x-csrftoken,X-CSRF-Token,Authorization,Accept,Origin,DNT,X-Mx-ReqToken,X-CustomHeader,Keep-Alive,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Content-Range,Range,Sec-Fetch-Mode,User-Agent' always;
"""  # noqa


def get_cors_config():
    return get_config(
        options=CORS_OPTIONS if settings.PROXIES_CONFIG.ui_single_url else "",
        indent=0,
    )



from typing import List, Optional

from clipped.utils.lists import to_list

from haupt.common import auditor
from polyaxon.schemas import V1StageCondition


def set_entity_stage(entity: str, condition: V1StageCondition) -> str:
    entity.stage = condition.type

    if condition:
        stage_conditions = None
        if entity.stage_conditions:
            stage_conditions = to_list(entity.stage_conditions, check_none=True)
            last_condition = V1StageCondition.get_condition(**stage_conditions[-1])
            if last_condition == condition:
                stage_conditions[-1] = condition.to_dict()
            else:
                stage_conditions.append(condition.to_dict())
        elif condition:
            stage_conditions = [condition.to_dict()]
        if stage_conditions:
            entity.stage_conditions = stage_conditions

    return entity


def new_stage(
    entity: str,
    condition: V1StageCondition,
    additional_fields: Optional[List[str]] = None,
    event_type: Optional[str] = None,
    user=None,
    owner_id: Optional[int] = None,
) -> str:
    previous_stage = entity.stage

    entity = set_entity_stage(entity=entity, condition=condition)

    additional_fields = additional_fields or []
    entity.save(
        update_fields=additional_fields
        + [
            "stage_conditions",
            "stage",
            "updated_at",
        ]
    )

    if event_type and user and owner_id:
        auditor.record(
            event_type=event_type,
            actor_id=user.id,
            actor_name=user.username,
            owner_id=owner_id,
            instance=entity,
            previous_stage=previous_stage,
        )

    return previous_stage



from typing import List, Tuple

from clipped.manager_interface import ManagerInterface

from haupt.common.events import event_actions
from haupt.common.events.event import Event


class EventManager(ManagerInterface):
    def _get_state_data(  # pylint:disable=arguments-differ
        self, event: Event
    ) -> Tuple[str, Event]:
        return event.event_type, event

    def subscribe(self, event: Event):  # pylint:disable=arguments-differ
        """
        >>> subscribe(SomeEvent)
        """
        super().subscribe(obj=event)

    def knows(self, event_type: str) -> bool:  # pylint:disable=arguments-differ
        return super().knows(key=event_type)

    def get(self, event_type: str) -> Event:  # pylint:disable=arguments-differ
        return super().get(key=event_type)

    def user_write_events(self) -> List[str]:
        """Return event types where use acted on an object.

        The write events are events with actions:
            * CREATED
            * UPDATED
            * DELETED
            * RESUMED
            * COPIED
            * CLONED
            * STOPPED
        """
        return [
            event_type
            for event_type, event in self.items
            if event.get_event_action() in event_actions.WRITE_ACTIONS
        ]

    def user_view_events(self) -> List[str]:
        """Return event types where use viewed a main object."""
        return [
            event_type
            for event_type, event in self.items
            if event.get_event_action() == event_actions.VIEWED
        ]



import sys

from typing import List

from clipped.formatting import Printer

from haupt.managers.sandbox import SandboxConfigManager
from haupt.settings import set_sandbox_config
from polyaxon._cli.config import set_home_path
from polyaxon._cli.errors import handle_cli_error
from polyaxon._managers.home import HomeConfigManager


def show():
    """Show the current sandbox config."""
    _config = HomeConfigManager.get_config_or_default()
    Printer.heading(
        "In addition to environment variables, global configs will be loaded from:"
    )
    Printer.dict_tabulate(_config.to_dict())
    _config = SandboxConfigManager.get_config_or_default()
    Printer.heading("Sandbox config:")
    Printer.dict_tabulate(_config.to_dict())


def get(keys: List[str]):
    """Get the specific keys from the sandbox configuration."""
    _config = SandboxConfigManager.get_config_or_default()

    if not keys:
        return

    print_values = {}
    for key in keys:
        key = key.replace("-", "_")
        if hasattr(_config, key):
            print_values[key] = getattr(_config, key)
        else:
            Printer.print("Key `{}` is not recognised.".format(key))

    Printer.dict_tabulate(print_values)


def set(**kwargs):  # pylint:disable=redefined-builtin
    """Set the sandbox config values."""
    try:
        _config = SandboxConfigManager.get_config_or_default()
    except Exception as e:
        handle_cli_error(e, message="Load configuration.")
        Printer.heading("You can reset your config by running: `sandbox config-purge`")
        sys.exit(1)

    for key, value in kwargs.items():
        if value is not None:
            if key == "path":
                set_home_path(home_path=value)
            else:
                setattr(_config, key, value)

    set_sandbox_config(_config, persist=True)
    Printer.success("Config was updated.")


def purge():
    """Purge the sandbox config values."""
    SandboxConfigManager.purge()
    Printer.success("Sandbox config was removed.")



import atexit
import logging
import threading
import time

from haupt.streams.connections.fs import AppFS
from polyaxon._constants.globals import DEFAULT, NO_AUTH
from polyaxon._fs.fs import get_sync_default_fs
from polyaxon._fs.manager import delete_file_or_dir
from polyaxon._fs.utils import get_store_path
from polyaxon._schemas.client import ClientConfig
from polyaxon.client import PolyaxonClient
from polyaxon.exceptions import ApiException
from urllib3.exceptions import HTTPError

_logger = logging.getLogger("haupt.cli.cron")


def agent_cron(host, stop_event: threading.Event):
    config = ClientConfig(host=host, token=NO_AUTH)
    polyaxon_client = PolyaxonClient(config=config)
    time.sleep(5)
    fs = get_sync_default_fs()
    store_path = AppFS.get_fs_root_path()
    while not stop_event.is_set():
        try:
            agent_state = polyaxon_client.agents_v1.cron_agent(
                owner=DEFAULT, _request_timeout=10, body={"state": True}
            )
            if agent_state and agent_state.state:
                deleting = agent_state.state.deleting or []
                for deleting_path in deleting:
                    subpath = get_store_path(
                        store_path=store_path, subpath=deleting_path
                    )
                    delete_file_or_dir(
                        fs=fs,
                        subpath=subpath,
                        is_file=False,
                    )
        except (ApiException, HTTPError) as e:
            _logger.debug(e)
            _logger.info("Waiting for Polyaxon to be ready...")
            time.sleep(5)
        time.sleep(2)


def start_cron(host: str):
    stop_event = threading.Event()
    thread = threading.Thread(
        target=agent_cron,
        args=(
            host,
            stop_event,
        ),
    )
    thread.daemon = True
    thread.start()

    # Ensure the thread stops gracefully on exit
    atexit.register(stop_event.set)



from typing import Dict

from haupt.background.celeryp.tasks import CronsCeleryTasks, SchedulerCeleryTasks


class TasksExecutions:
    _MAPPING = None
    _MANAGER = None

    @classmethod
    def get_mapping(cls):
        if cls._MAPPING is None:
            cls._set_mapping()
        return cls._MAPPING

    @classmethod
    def get_manager(cls):
        if cls._MANAGER is None:
            cls._set_manager()
        return cls._MANAGER

    @classmethod
    def run(cls, task: str, kwargs: Dict = None, eager_kwargs: Dict = None) -> Dict:
        kwargs = kwargs or {}
        eager_kwargs = eager_kwargs or {}
        mapping = cls.get_mapping()
        mapping[task](**kwargs, **eager_kwargs)

    @classmethod
    def register_tasks(cls):
        from haupt.common import workers

        mapping = cls.get_mapping()
        for task in mapping:
            workers.app.task(name=task, ignore_result=True)(mapping[task])

    @classmethod
    def _set_manager(cls):
        from haupt.background.celeryp.manager import BackgroundManager

        cls._MANAGER = BackgroundManager

    @classmethod
    def _set_mapping(cls):
        from haupt.common.checks import health_task

        manager = cls.get_manager()

        cls._MAPPING = {
            SchedulerCeleryTasks.STATS_CALCULATION_PROJECT: manager.stats_calculation_project,
            SchedulerCeleryTasks.SCHEDULER_HEALTH: health_task.health_task,
            SchedulerCeleryTasks.RUNS_PREPARE: manager.runs_prepare,
            SchedulerCeleryTasks.RUNS_HOOKS: manager.runs_hooks,
            SchedulerCeleryTasks.RUNS_START_IMMEDIATELY: manager.runs_start_immediately,
            SchedulerCeleryTasks.RUNS_START: manager.runs_start,
            SchedulerCeleryTasks.RUNS_BUILT: manager.runs_built,
            SchedulerCeleryTasks.RUNS_STOP: manager.runs_stop,
            SchedulerCeleryTasks.RUNS_SET_ARTIFACTS: manager.runs_set_artifacts,
            SchedulerCeleryTasks.RUNS_NOTIFY_STATUS: manager.runs_notify_status,
            SchedulerCeleryTasks.RUNS_NOTIFY_DONE: manager.runs_notify_done,
            SchedulerCeleryTasks.RUNS_CHECK_PIPELINE: manager.runs_check_pipeline,
            SchedulerCeleryTasks.RUNS_CHECK_ORPHAN_PIPELINE: manager.runs_check_orphan_pipeline,
            SchedulerCeleryTasks.RUNS_CHECK_EARLY_STOPPING: manager.runs_check_early_stopping,
            SchedulerCeleryTasks.RUNS_WAKEUP_SCHEDULE: manager.runs_wakeup_schedule,
            SchedulerCeleryTasks.RUNS_ITERATE: manager.runs_iterate,
            SchedulerCeleryTasks.RUNS_TUNE: manager.runs_tune,
            SchedulerCeleryTasks.DELETE_ARCHIVED_PROJECT: manager.delete_archived_project,
            SchedulerCeleryTasks.DELETE_ARCHIVED_RUN: manager.delete_archived_run,
            CronsCeleryTasks.CRONS_HEALTH: health_task.health_task,
            CronsCeleryTasks.DELETE_ARCHIVED_PROJECTS: manager.delete_archived_projects,
            CronsCeleryTasks.DELETE_IN_PROGRESS_PROJECTS: manager.delete_in_progress_projects,
            CronsCeleryTasks.DELETE_ARCHIVED_RUNS: manager.delete_archived_runs,
            CronsCeleryTasks.DELETE_IN_PROGRESS_RUNS: manager.delete_in_progress_runs,
            CronsCeleryTasks.HEARTBEAT_OUT_OF_SYNC_SCHEDULES: manager.heartbeat_out_of_sync_schedules,
            CronsCeleryTasks.HEARTBEAT_STOPPING_RUNS: manager.heartbeat_stopping_runs,
            CronsCeleryTasks.HEARTBEAT_PROJECT_LAST_UPDATED: manager.heartbeat_project_last_updated,
            CronsCeleryTasks.STATS_CALCULATION_PROJECTS: manager.stats_calculation_projects,
        }



import click

from haupt.cli.manage import manage
from haupt.cli.proxies import proxy
from haupt.cli.queues import queues
from haupt.cli.sandbox import sandbox
from haupt.cli.server import server
from haupt.cli.streams import streams
from haupt.cli.viewer import viewer
from polyaxon.logger import clean_outputs, configure_logger


@click.group()
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Turn on debug logging"
)
@clean_outputs
def cli(verbose):
    """Haupt -  Lineage metadata API, artifacts streams, sandbox, ML-API, and spaces for Polyaxon.

    Check the help available for each command listed below by appending `-h`.
    """
    configure_logger(verbose)


cli.add_command(proxy)
cli.add_command(manage)
cli.add_command(sandbox)
cli.add_command(server)
cli.add_command(streams)
cli.add_command(viewer)
cli.add_command(queues)



from haupt import settings
from haupt.proxies.schemas.base import get_config

OPTIONS = """
error_log {root}/error.log {level};
"""


def get_logging_config():
    return get_config(
        options=OPTIONS,
        indent=0,
        root=settings.PROXIES_CONFIG.logs_root,
        level=settings.PROXIES_CONFIG.get_log_level(),
    )



import json

from typing import List

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder

from haupt.common.db.inserter import RawBulkInserter
from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models

_RUN_FIELDS = (
    "description",
    "created_at",
    "updated_at",
    "live_state",
    "tags",
    "uuid",
    "started_at",
    "finished_at",
    "wait_time",
    "duration",
    "status",
    "status_conditions",
    "state",
    "readme",
    "component_state",
    "raw_content",
    "content",
    "name",
    "kind",
    "runtime",
    "project_id",
    "managed_by",
    "pending",
    "meta_info",
    "params",
    "inputs",
    "outputs",
    "original_id",
    "pipeline_id",
    "cloning_kind",
    "controller_id",
    "schedule_at",
    "user_id",
)

if settings.HAS_ORG_MANAGEMENT:
    _RUN_FIELDS += (
        "namespace",
        "artifacts_store_id",
        "agent_id",
        "queue_id",
    )


def _process_field(run: BaseRun, field: str):
    value = getattr(run, field)
    if field in {"status_conditions", "meta_info", "inputs", "outputs", "params"}:
        return json.dumps(value, cls=DjangoJSONEncoder)
    return value


def bulk_create_runs(runs: List[BaseRun], fetch_ids: bool = False):
    if settings.DB_ENGINE_NAME == "sqlite":
        Models.Run.objects.bulk_create(runs)
        return
    inserter = RawBulkInserter(Models.Run, _RUN_FIELDS, fetch_ids=fetch_ids)
    for run in runs:
        inserter.queue_row(*[_process_field(run, i) for i in _RUN_FIELDS])

    inserter.insert()



from collections.abc import Mapping
from typing import Dict, Optional

import click


def sanitize_server_config(server_config: Optional[Dict] = None) -> Dict:
    from clipped.formatting import Printer

    if not server_config:
        return {}

    def parse_server_config():
        if isinstance(parse_server_config, Mapping):
            return parse_server_config

        parsed_server_config = {}
        for sc in server_config:
            index = sc.find("=")
            if index == -1:
                message = (
                    "Invalid format for -SC/--server-config option: '%s'. Use -SC name=value."
                    % sc
                )
                Printer.error(message, sys_exit=True)
            name = sc[:index]
            value = sc[index + 1 :]
            if name in parsed_server_config:
                message = "Repeated parameter: '%s'" % name
                Printer.error(message, sys_exit=True)
            parsed_server_config[name] = value

        return parsed_server_config

    server_config = parse_server_config()
    keys = {
        "host",
        "port",
        "workers",
        "per_core",
        "per-core",
        "path",
        "uds",
    }
    return {
        key: server_config[key] for key in keys if server_config.get(key) is not None
    }


@click.command()
@click.option(
    "--host",
    help="The service host.",
)
@click.option(
    "--port",
    type=int,
    help="The service port.",
)
@click.option(
    "--workers",
    type=int,
    help="Number of workers.",
)
@click.option(
    "--per-core",
    is_flag=True,
    default=False,
    help="To enable workers per core.",
)
@click.option(
    "--path",
    help="The sandbox root path.",
)
@click.option(
    "--uds",
    help="UNIX domain socket binding.",
)
def viewer(
    host: str,
    port: int,
    workers: int,
    per_core: bool,
    path: str,
    uds: str,
):
    """Start a new local viewer session."""
    from haupt.cli.runners.viewer import start

    return start(
        host=host, port=port, workers=workers, per_core=per_core, uds=uds, path=path
    )



import click

from polyaxon.logger import clean_outputs


@click.group()
@clean_outputs
def sandbox():
    """Command for sandbox."""


@sandbox.command()
@click.option(
    "--host",
    help="The service host.",
)
@click.option(
    "--port",
    type=int,
    help="The service port.",
)
@click.option(
    "--workers",
    type=int,
    help="Number of workers.",
)
@click.option(
    "--per-core",
    is_flag=True,
    default=False,
    help="To enable workers per core.",
)
@click.option(
    "--path",
    help="The sandbox root path.",
)
@click.option(
    "--uds",
    help="UNIX domain socket binding.",
)
def start(
    host: str,
    port: int,
    workers: int,
    per_core: bool,
    path: str,
    uds: str,
):
    """Start a new sandbox session."""
    from haupt.cli.runners.sandbox import start
    from polyaxon._cli.config import set_home_path

    path = path or "."
    set_home_path(home_path=path)

    return start(host=host, port=port, workers=workers, per_core=per_core, uds=uds)


@sandbox.command()
@clean_outputs
def show_config():
    """Show the current sandbox config."""
    from haupt.cli.runners.config import show

    return show()


@sandbox.command()
@clean_outputs
@click.option(
    "--host",
    help="The service host.",
)
@click.option(
    "--port",
    type=int,
    help="The service port.",
)
@click.option(
    "--workers",
    type=int,
    help="Number of workers.",
)
@click.option(
    "--per-core",
    is_flag=True,
    default=False,
    help="To enable workers per core.",
)
@click.option(
    "--path",
    help="The sandbox root path.",
)
def set_config(
    host: str,
    port: int,
    workers: int,
    per_core: bool,
    path: str,
):
    """Set the current sandbox config."""
    from haupt.cli.runners.config import set

    return set(
        host=host,
        port=port,
        workers=workers,
        per_core=per_core,
        path=path,
    )


@sandbox.command()
@clean_outputs
def purge_config():
    """Purge the current sandbox config."""
    from haupt.cli.runners.config import purge

    return purge()



from typing import Any, Optional, Tuple

from rest_framework import HTTP_HEADER_ENCODING, exceptions
from rest_framework.authentication import get_authorization_header

from django.http import HttpRequest

from haupt.common import conf
from haupt.common.authentication.base import BaseUser, PolyaxonAuthentication
from haupt.common.headers import get_internal_header
from haupt.common.options.registry.core import SECRET_INTERNAL_TOKEN
from polyaxon._services.auth import AuthenticationTypes
from polyaxon._services.values import PolyaxonServices


class InternalUser(BaseUser):
    def __init__(self, service: str):
        self.username = "internal_user"
        self.is_internal = True
        super().__init__(service)

    @property
    def access_token(self) -> str:
        return conf.get(SECRET_INTERNAL_TOKEN)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, InternalUser) and other.username == self.username


def is_internal_user(user: Any) -> bool:
    return hasattr(user, "is_internal")


def is_authenticated_internal_user(user: Any) -> bool:
    if is_internal_user(user):
        return user.is_internal

    return False


class InternalAuthentication(PolyaxonAuthentication):
    """
    Simple authentication based on internal secret token.

    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string "InternalToken ".  For example:

        Authorization: InternalToken 401f7ac837da42b97f613d789819ff93537bee6a

    As well as one of the supported internal service. For example:

        X_POLYAXON_INTERNAL: experiments
    """

    keyword = AuthenticationTypes.INTERNAL_TOKEN

    def authenticate(
        self, request: HttpRequest
    ) -> Optional[Tuple["InternalUser", None]]:
        data = get_internal_auth(request=request, keyword=self.keyword)

        if data is None:
            return None

        internal_service, token = data
        return self.authenticate_credentials(internal_service, token)

    def authenticate_credentials(
        self,
        service: str,
        key: str,  # pylint:disable=arguments-differ
    ) -> Optional[Tuple["InternalUser", None]]:
        internal_user = InternalUser(service=service)
        if internal_user.access_token != key:
            raise exceptions.AuthenticationFailed("Invalid token.")

        return internal_user, None


def get_internal_auth(
    request: HttpRequest, keyword: str = AuthenticationTypes.INTERNAL_TOKEN
) -> Optional[Tuple[str, str]]:
    auth = get_authorization_header(request).split()
    internal_service = get_internal_header(request)
    try:
        internal_service = internal_service.decode()
    except UnicodeError:
        msg = (
            "Invalid internal_service header. "
            "internal_service string should not contain invalid characters."
        )
        raise exceptions.AuthenticationFailed(msg)

    if internal_service not in PolyaxonServices.to_set():
        return None

    if not auth or auth[0].lower() != keyword.lower().encode(HTTP_HEADER_ENCODING):
        return None

    if len(auth) == 1:
        msg = "Invalid token header. No credentials provided."
        raise exceptions.AuthenticationFailed(msg)

    elif len(auth) > 2:
        msg = "Invalid token header. Token string should not contain spaces."
        raise exceptions.AuthenticationFailed(msg)

    try:
        token = auth[1].decode()
    except UnicodeError:
        msg = (
            "Invalid token header. Token string should not contain invalid characters."
        )
        raise exceptions.AuthenticationFailed(msg)

    return internal_service, token



import time
import uuid

from functools import reduce
from operator import or_
from typing import Any, Dict, List, Set

from django.conf import settings
from django.db import IntegrityError, transaction
from django.db.models import Q

from haupt.db.abstracts.projects import Owner
from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models
from traceml.artifacts import V1ArtifactKind, V1RunArtifact


def get_artifacts_by_keys(
    run: BaseRun, namespace: uuid.UUID, artifacts: List[V1RunArtifact]
) -> Dict:
    results = {}
    for m in artifacts:
        state = m.state
        if not state:
            if m.is_input:
                state = m.get_state(namespace)
            else:
                state = run.uuid
        elif not isinstance(state, uuid.UUID):
            try:
                state = uuid.UUID(state)
            except (ValueError, KeyError):
                state = uuid.uuid5(namespace, state)
        results[(m.name, state)] = m

    return results


def set_run_lineage(run: BaseRun, artifacts_by_keys: Dict, query: Any):
    artifacts_to_link = Models.Artifact.objects.filter(query).only(
        "id", "name", "state"
    )
    for m in artifacts_to_link:
        Models.ArtifactLineage.objects.get_or_create(
            artifact_id=m.id,
            run_id=run.id,
            is_input=artifacts_by_keys[(m.name, m.state)].is_input,
        )


def update_artifacts(to_update: Set, artifacts_by_keys: Dict):
    updated = []
    for m in to_update:
        artifact = artifacts_by_keys[(m.name, m.state)]
        m.kind = artifact.kind
        m.path = artifact.path
        m.summary = artifact.summary
        updated.append(m)
    Models.Artifact.objects.bulk_update(updated, ["kind", "path", "summary"])


def set_artifacts(run: BaseRun, artifacts: List[V1RunArtifact]):
    if not artifacts:
        return

    owner = run.project.owner
    namespace = owner.uuid if settings.HAS_ORG_MANAGEMENT else Owner.uuid

    artifacts_by_keys = get_artifacts_by_keys(
        run=run, namespace=namespace, artifacts=artifacts
    )
    artifacts_keys = list(artifacts_by_keys.keys())
    query = reduce(or_, (Q(name=name, state=state) for name, state in artifacts_keys))
    if settings.HAS_ORG_MANAGEMENT:
        query = Q(owner=owner) & query
    to_update = Models.Artifact.objects.filter(query)
    to_create = {
        m for m in artifacts_keys if m not in {(n.name, n.state) for n in to_update}
    }

    has_iteration_kind = False
    has_metrics = False
    owner_filter = {"owner": owner} if settings.HAS_ORG_MANAGEMENT else {}
    if to_create:
        artifacts_to_create = []
        for m in to_create:
            a = artifacts_by_keys[m]
            if a.kind == V1ArtifactKind.METRIC and not a.is_input:
                has_metrics = True
            if a.kind == V1ArtifactKind.ITERATION and not a.is_input:
                has_iteration_kind = True
            artifacts_to_create.append(
                Models.Artifact(
                    **owner_filter,
                    name=a.name,
                    kind=a.kind,
                    path=a.path,
                    state=m[1],
                    summary=a.summary,
                )
            )
        Models.Artifact.objects.bulk_create(artifacts_to_create)

    update_artifacts(to_update=to_update, artifacts_by_keys=artifacts_by_keys)
    set_run_lineage(run=run, artifacts_by_keys=artifacts_by_keys, query=query)
    # Handle iteration lineage
    # if has_iteration_kind:
    #     workers.send(SchedulerCeleryTasks.RUNS_ITERATE, kwargs={"run_id": run.id})
    # TODO: Add intermediate metric early


@transaction.atomic
def atomic_set_artifacts(run: BaseRun, artifacts: List[V1RunArtifact]):
    retries = 0
    while retries < 2:
        try:
            return set_artifacts(run=run, artifacts=artifacts)
        except IntegrityError:
            retries += 1
            time.sleep(0.01)



from haupt import settings


def _get_indent(indent):
    return (
        settings.PROXIES_CONFIG.nginx_indent_char
        * settings.PROXIES_CONFIG.nginx_indent_width
        * indent
    )


def get_config(options, indent=0, **kwargs):
    _options = options.format(**kwargs)
    config = []
    for p in _options.split("\n"):
        config.append("{}{}".format(_get_indent(indent), p))

    return clean_config(config)


def clean_config(config):
    return "\n".join(config).replace("    \n", "")



from haupt.db.defs import Models


def remove_bookmarks(object_id: int, content_type: str):
    # Remove any bookmark
    Models.Bookmark.objects.filter(
        content_type__model=content_type, object_id=object_id
    ).delete()


def bookmark_obj(user, obj, content_type: str):
    user_filter = {"user": user} if user else {}
    try:
        bookmark = Models.Bookmark.objects.get(
            content_type__model=content_type, object_id=obj.id, **user_filter
        )
        bookmark.enabled = True
        bookmark.save(update_fields=["enabled"])
    except Models.Bookmark.DoesNotExist:
        Models.Bookmark.objects.create(content_object=obj, **user_filter)



import logging
import os

from typing import Dict, Optional

from clipped.utils.dates import DateTimeFormatter
from clipped.utils.hashing import hash_value
from clipped.utils.json import orjson_dumps, orjson_loads
from clipped.utils.paths import check_or_create_path
from clipped.utils.tz import now
from rest_framework import status

from django.conf import settings as dj_settings
from django.core.handlers.asgi import ASGIRequest
from django.db import transaction
from django.http import HttpResponse
from django.urls import path

import aiofiles
import aiohttp

from haupt.common.endpoints.validation import validate_methods
from haupt.common.headers import (
    get_authorization_header,
    get_original_method_header,
    get_original_uri_header,
)
from haupt.streams.controllers.k8s_check import k8s_check, reverse_k8s
from polyaxon import settings
from polyaxon._env_vars.keys import ENV_KEYS_PROXY_LOCAL_PORT
from polyaxon.api import AUTH_V1_LOCATION

logger = logging.getLogger("haupt.streams.auth")


def _get_auth_cache_path(request_cache: str) -> str:
    auth_cache_path = os.path.join(
        settings.CLIENT_CONFIG.archives_root, ".token/{}".format(request_cache)
    )
    check_or_create_path(auth_cache_path, is_dir=False)
    return auth_cache_path


async def _check_auth_cache(request_cache: str) -> bool:
    cache_path = _get_auth_cache_path(request_cache)
    if os.path.isfile(cache_path):
        try:
            async with aiofiles.open(cache_path, mode="r") as f:
                last_value = orjson_loads(await f.read())
                last_time = DateTimeFormatter.extract_timestamp(
                    last_value["time"],
                    dt_format=DateTimeFormatter.DATETIME_FORMAT,
                    timezone=settings.CLIENT_CONFIG.timezone,
                    force_tz=True,
                )
                interval = 60 if last_value["response"] else 30

            if (now() - last_time).seconds < interval:
                return True

            os.remove(cache_path)
        except Exception as e:
            logger.warning(
                "An error was raised while deleting some auth cache request %s" % e
            )

    return False


async def _persist_auth_cache(request_cache: str, response: bool):
    cache_path = _get_auth_cache_path(request_cache)
    data = {"time": DateTimeFormatter.format_datetime(now()), "response": response}
    async with aiofiles.open(cache_path, "w") as outfile:
        await outfile.write(orjson_dumps(data))


async def _check_auth_service(headers: Dict):
    async with aiohttp.ClientSession(trust_env=True) as session:
        async with session.get(
            "http://localhost:{}{}".format(
                os.environ[ENV_KEYS_PROXY_LOCAL_PORT], AUTH_V1_LOCATION
            ),
            headers=headers,
        ) as resp:
            return resp


@transaction.non_atomic_requests
async def auth_request(
    request: ASGIRequest, methods: Optional[Dict] = None
) -> HttpResponse:
    validate_methods(request, methods)
    # Polyaxon checks for origin headers
    try:
        uri = get_original_uri_header(request)
        logger.debug("Authenticating %s" % uri)
        if not uri:
            return HttpResponse(
                content="Request must comply with an HTTP_X_ORIGIN_URI",
                status=status.HTTP_403_FORBIDDEN,
            )
        method = get_original_method_header(request)
        if not method:
            return HttpResponse(
                content="Request must comply with an HTTP_X_ORIGIN_METHOD",
                status=status.HTTP_403_FORBIDDEN,
            )
        auth = get_authorization_header(request)
        request_cache = hash_value(
            value={uri.split("?")[0], method, auth}, hash_length=64
        )
    except Exception as exc:
        return HttpResponse(
            content="Auth request failed extracting headers: %s" % exc,
            status=status.HTTP_403_FORBIDDEN,
        )

    # Check cache
    cached = await _check_auth_cache(request_cache=request_cache)
    if cached:
        return HttpResponse(status=status.HTTP_200_OK)

    if dj_settings.POLYAXON_SERVICE == "streams":
        # Contact auth service
        response = await _check_auth_service(headers=request.headers)
        if response.status != 200:
            await _persist_auth_cache(request_cache=request_cache, response=False)
            return HttpResponse(
                content="Auth request failed",
                status=status.HTTP_403_FORBIDDEN,
            )
    try:
        path, params = k8s_check(uri)
    except ValueError:
        await _persist_auth_cache(request_cache=request_cache, response=True)
        return HttpResponse(status=status.HTTP_200_OK)
    return await reverse_k8s(path="{}?{}".format(path, params))


URLS_RUNS_AUTH_REQUEST = ""

# fmt: off
auth_request_routes = [
    path(
        URLS_RUNS_AUTH_REQUEST,
        auth_request,
        name="auth_request",
        kwargs=dict(methods=["GET"]),
    ),
]



from django.conf import settings
from django.core.validators import validate_slug
from django.db import models

from haupt.db.abstracts.contributors import ContributorsModel
from haupt.db.abstracts.describable import DescribableModel
from haupt.db.abstracts.diff import DiffModel
from haupt.db.abstracts.duration import DurationModel
from haupt.db.abstracts.live_state import LiveStateModel
from haupt.db.abstracts.nameable import NameableModel
from haupt.db.abstracts.readme import ReadmeModel
from haupt.db.abstracts.run_pipelines import RunPipelines
from haupt.db.abstracts.run_resources import RunResources
from haupt.db.abstracts.spec import SpecModel
from haupt.db.abstracts.state import OptionalStateModel
from haupt.db.abstracts.status import StatusModel
from haupt.db.abstracts.tag import TagModel
from haupt.db.abstracts.uid import UuidModel
from haupt.db.defs import Models
from polyaxon.schemas import (
    ManagedBy,
    V1CloningKind,
    V1RunKind,
    V1RunPending,
    V1Statuses,
)


class BaseRun(
    UuidModel,
    DiffModel,
    DurationModel,
    SpecModel,
    NameableModel,
    DescribableModel,
    ReadmeModel,
    StatusModel,
    TagModel,
    LiveStateModel,
    OptionalStateModel,
    RunPipelines,
    RunResources,
    ContributorsModel,
):
    name = models.CharField(
        max_length=128,
        blank=True,
        null=True,
        default=None,
        validators=[validate_slug],
    )
    kind = models.CharField(
        max_length=12,
        db_index=True,
        choices=V1RunKind.to_choices(),
    )
    runtime = models.CharField(max_length=12, db_index=True, null=True, blank=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="+",
        null=True,
        blank=True,
    )
    project = models.ForeignKey(
        Models.get_db_model_name("Project"),
        on_delete=models.CASCADE,
        related_name="runs",
    )
    managed_by = models.CharField(
        max_length=5,
        null=True,
        blank=True,
        default=ManagedBy.AGENT,
        choices=ManagedBy.to_choices(),
        db_index=True,
    )
    pending = models.CharField(
        max_length=8,
        null=True,
        blank=True,
        db_index=True,
        choices=V1RunPending.to_choices(),
        help_text="If this entity requires approval before it should run.",
    )
    meta_info = models.JSONField(null=True, blank=True, default=dict)
    params = models.JSONField(null=True, blank=True)
    inputs = models.JSONField(null=True, blank=True)
    outputs = models.JSONField(null=True, blank=True)
    component_state = models.UUIDField(null=True, blank=True, db_index=True)
    original = models.ForeignKey(
        "self",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="clones",
    )
    cloning_kind = models.CharField(
        max_length=12,
        blank=True,
        null=True,
        choices=V1CloningKind.to_choices(),
    )
    artifacts = models.ManyToManyField(
        Models.get_db_model_name("Artifact"),
        blank=True,
        through=Models.get_db_model_name("ArtifactLineage"),
        related_name="runs",
    )

    class Meta:
        abstract = True
        app_label = "db"
        db_table = "db_run"
        indexes = [models.Index(fields=["name"])]

    @property
    def subpath(self):
        return self.uuid.hex

    @property
    def is_managed(self) -> bool:
        return ManagedBy.is_managed(self.managed_by)

    @property
    def is_clone(self) -> bool:
        return self.original_id is not None

    @property
    def is_restart(self) -> bool:
        return self.is_clone and self.cloning_kind == V1CloningKind.RESTART

    @property
    def is_resume(self) -> bool:
        if not self.status_conditions:
            return False
        return bool(
            [c for c in self.status_conditions if c.get("type") == V1Statuses.RESUMING]
        )

    @property
    def is_copy(self) -> bool:
        return self.is_clone and self.cloning_kind == V1CloningKind.COPY

    @property
    def is_job(self):
        return self.kind == V1RunKind.JOB

    @property
    def is_service(self):
        return self.kind == V1RunKind.SERVICE

    @property
    def is_dag(self):
        return self.kind == V1RunKind.DAG

    @property
    def is_matrix(self):
        return self.kind == V1RunKind.MATRIX

    @property
    def is_schedule(self):
        return self.kind == V1RunKind.SCHEDULE

    @property
    def has_pipeline(self):
        return self.is_dag or self.is_matrix or self.is_schedule

    @property
    def has_tuner_runtime(self):
        return self.runtime == V1RunKind.TUNER

    @property
    def has_notifier_runtime(self):
        return self.runtime == V1RunKind.NOTIFIER

    def get_last_condition(self):
        return (self.status_conditions or [{}])[-1]



from django.conf import settings
from django.db.models import Q

from polyaxon._pql.manager import PQLManager


class BaseQueryManager(PQLManager):
    QUERY_BACKEND = Q
    TIMEZONE = settings.TIME_ZONE



import random

from django.db import models
from django.utils.timezone import now

from haupt.db.managers.deleted import ArchivedManager, LiveManager, RestorableManager
from polyaxon.schemas import LiveState


class LiveStateModel(models.Model):
    live_state = models.IntegerField(
        null=True,
        blank=True,
        default=LiveState.LIVE,
        choices=LiveState.to_choices(),
        db_index=True,
    )
    deleted_at = models.DateTimeField(null=True, blank=True)
    archived_at = models.DateTimeField(null=True, blank=True)

    objects = LiveManager()
    restorable = RestorableManager()
    all = models.Manager()
    archived = ArchivedManager()

    class Meta:
        abstract = True

    def delete_in_progress(
        self,
        update_name: bool = False,
        set_deleted_at: bool = True,
        commit: bool = True,
    ) -> bool:
        if self.live_state == LiveState.DELETION_PROGRESSING:
            return False

        self.live_state = LiveState.DELETION_PROGRESSING
        update_fields = ["live_state", "updated_at"]
        if update_name:
            self.name = "del_{}_{}".format(
                self.name, getattr(self, "uuid", random.randint(-90, 100))
            )
            update_fields.append("name")
        if set_deleted_at:
            self.deleted_at = now()
            update_fields.append("deleted_at")
        if commit:
            self.save(update_fields=update_fields)
        return True

    def confirm_delete(self) -> bool:
        if self.deleted_at:
            return False

        update_fields = ["deleted_at", "updated_at"]
        self.deleted_at = now()
        if self.live_state != LiveState.DELETION_PROGRESSING:
            self.live_state = LiveState.DELETION_PROGRESSING
            update_fields.append("live_state")
        self.save(update_fields=update_fields)
        return True

    def archive(self, commit: bool = True) -> bool:
        if self.live_state in {
            LiveState.ARCHIVED,
            LiveState.DELETION_PROGRESSING,
        }:
            return False

        self.archived_at = now()
        self.live_state = LiveState.ARCHIVED
        if commit:
            self.save(update_fields=["live_state", "archived_at", "updated_at"])
        return True

    def restore(self) -> bool:
        if self.live_state != LiveState.ARCHIVED:
            return False

        self.live_state = LiveState.LIVE
        update_fields = ["live_state", "updated_at"]
        if self.archived_at:
            self.archived_at = None
            update_fields.append("archived_at")
        if self.deleted_at:
            self.deleted_at = None
            update_fields.append("deleted_at")
        self.save(update_fields=update_fields)
        return True



RUN = "run"
ACTOR = "actor"
PROJECT = "project"
REPO = "repo"
USER = "user"
BOOKMARK = "bookmark"
ARCHIVE = "archive"
PROJECT_SEARCH = "project_search"
ORGANIZATION = "organization"
ORGANIZATION_SEARCH = "organization_search"
PROJECT_DASHBOARD = "project_dashboard"
ORGANIZATION_DASHBOARD = "organization_dashboard"
TAG = "tag"
WORKER = "worker"
AGENT = "agent"
COMPONENT_VERSION = "component_version"
MODEL_VERSION = "model_version"
ARTIFACT_VERSION = "artifact_version"
PRESET = "preset"
QUEUE = "queue"
CONNECTION = "connection"
MEMBER = "member"
INVITATION = "invitation"
TEAM = "team"
SERVICE_ACCOUNT = "service_account"
SERVICE_ACCOUNT_TOKEN = "service_account_token"
POLICY = "policy"



from typing import Any, Iterable, Optional, Union

from clipped.utils.bools import to_bool
from clipped.utils.lists import to_list

from polyaxon._schemas.lifecycle import V1ProjectFeature
from polyaxon.schemas import LiveState, ManagedBy
from traceml.artifacts import V1ArtifactKind


def archived_condition(
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    queryset: Any = None,
    request: Optional[Any] = None,
) -> Any:
    """
    Example:
        >>>  {"archived": CallbackCondition(callback_conditions.archived_condition)}
    """
    params = to_list(params)
    if len(params) == 1 and to_bool(params[0]) is True:
        return (
            queryset.filter(live_state=LiveState.ARCHIVED)
            if queryset
            else query_backend(live_state=LiveState.ARCHIVED)
        )
    return query_backend(live_state=LiveState.LIVE)


def independent_condition(
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    queryset: Any = None,
    request: Optional[Any] = None,
) -> Any:
    params = to_list(params)
    if len(params) == 1 and to_bool(params[0]) is True:
        return query_backend(experiment_group__isnull=True)
    return None


def metric_condition(
    queryset: Any,
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    request: Optional[Any] = None,
) -> Any:
    params = to_list(params)
    if len(params) == 1 and to_bool(params[0]) is True:
        return query_backend(metric_annotations__name=True)
    return None


def commit_condition(
    queryset: Any,
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    request: Optional[Any] = None,
) -> Any:
    params = to_list(params)
    if len(params) == 1:
        if negation:
            return ~query_backend(
                artifacts__kind=V1ArtifactKind.CODEREF,
                artifacts__name=params[0],
            )
        return query_backend(
            artifacts__kind=V1ArtifactKind.CODEREF,
            artifacts__name=params[0],
        )
    return None


def _artifact_kind_condition(
    is_input: bool,
    query_backend: Any,
    params: Union[str, Iterable],
    negation: bool,
) -> Any:
    params = to_list(params)
    if len(params) == 1:
        if negation:
            return ~query_backend(
                artifacts_lineage__is_input=is_input,
                artifacts__kind=params[0],
            )
        return query_backend(
            artifacts_lineage__is_input=is_input,
            artifacts__kind=params[0],
        )
    return None


def in_artifact_kind_condition(
    queryset: Any,
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    request: Optional[Any] = None,
) -> Any:
    return _artifact_kind_condition(True, query_backend, params, negation)


def out_artifact_kind_condition(
    queryset: Any,
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    request: Optional[Any] = None,
) -> Any:
    return _artifact_kind_condition(False, query_backend, params, negation)


def is_managed_condition(
    queryset: Any,
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    request: Optional[Any] = None,
) -> Any:
    params = to_list(params)
    if len(params) == 1:
        if (to_bool(params[0]) is True and not negation) or (
            to_bool(params[0]) is False and negation
        ):
            return ~query_backend(managed_by=ManagedBy.USER)
        else:
            return query_backend(managed_by=ManagedBy.USER)
    return None


def mine_condition(
    queryset: Any,
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    request: Optional[Any] = None,
) -> Any:
    params = to_list(params)
    if len(params) == 1:
        query = query_backend(user=request.user)
        if negation:
            return ~query
        return query
    return None


def project_kind_condition(
    queryset: Any,
    params: Union[str, Iterable],
    negation: bool,
    query_backend: Any,
    timezone: Optional[str] = None,
    request: Optional[Any] = None,
) -> Any:
    params = to_list(params)
    if len(params) == 1:
        kind = params[0]
        project_features = V1ProjectFeature.to_set()
        if kind not in project_features:
            return None
        excluded_features = list(project_features - {kind})
        query = query_backend(excluded_features__contains=excluded_features)
        if negation:
            return ~query
        return query
    return None



import factory

from haupt.db.defs import Models


class ProjectFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence("project-{}".format)

    class Meta:
        model = Models.Project


class ProjectVersionFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence("artifact-version-{}".format)

    class Meta:
        model = Models.ProjectVersion



import copy
import datetime as _datetime

from typing import Any, Dict, Iterable, Mapping, Optional, Union
from uuid import UUID, uuid1

from clipped.utils.dates import to_datetime, to_timestamp

from django.conf import settings
from django.db.models import Model
from django.utils import timezone

from haupt.common import user_system
from haupt.common.events import event_context
from haupt.common.json_utils import dumps_htmlsafe


class Attribute:
    def __init__(
        self,
        name: str,
        attr_type: Any = str,
        is_datetime: bool = False,
        is_uuid: bool = False,
        is_required: bool = True,
    ):
        assert name != "instance"
        self.name = name
        self.attr_type = attr_type
        self.is_datetime = is_datetime
        self.is_uuid = is_uuid
        self.is_required = is_required

    def extract(self, value: Any) -> Any:
        if value is None:
            return value
        if self.is_datetime:
            return to_timestamp(value)
        if self.is_uuid and not isinstance(value, str):
            return value.hex
        return self.attr_type(value)


class Event:
    __slots__ = [
        "uuid",
        "ref_id",
        "data",
        "datetime",
        "instance",
        "instance_id",
        "instance_uuid",
    ]

    event_type = None  # The event type should ideally follow subject.action
    attributes = ()
    actor = False
    actor_id = "actor_id"
    actor_name = "actor_name"
    entity_uuid = None
    owner_id = None
    owner_name = None

    @classmethod
    def get_event_attributes(cls) -> Iterable["Attribute"]:
        attributes = cls.attributes + (
            Attribute("ref_id", is_uuid=True, is_required=False),
        )
        if cls.actor:
            return attributes + (
                Attribute(
                    cls.actor_id, attr_type=int, is_required=settings.HAS_ORG_MANAGEMENT
                ),
                Attribute(cls.actor_name, is_required=False),
            )
        return attributes

    def __init__(
        self,
        uid: Optional[str] = None,
        datetime: _datetime.datetime = None,
        instance: Any = None,
        instance_id: Optional[int] = None,
        instance_uuid: Optional[str] = None,
        ref_id: Optional[str] = None,
        event_data: Mapping = None,
        **items,
    ):
        self.uuid = UUID(uid) if uid else uuid1()
        self.datetime = datetime or timezone.now()
        self.instance = instance
        self.instance_id = instance_id
        self.instance_uuid = instance_uuid
        self.ref_id = None
        if ref_id:
            self.ref_id = UUID(ref_id) if isinstance(ref_id, str) else ref_id

        if self.event_type is None:
            raise ValueError("Event is missing a type")

        if event_data:
            self.data = event_data
        else:
            data = {}
            for attr in self.get_event_attributes():
                # Check plain attr name
                item_value = items.pop(attr.name, None)
                if item_value is None:
                    # Convert dot notation
                    item_value = items.pop(attr.name.replace(".", "_"), None)

                if attr.is_required and item_value is None:
                    raise ValueError(
                        "{} is required (cannot be None)".format(attr.name)
                    )
                data[attr.name] = attr.extract(item_value)

            if self.actor and not settings.HAS_ORG_MANAGEMENT:
                data[self.actor_id] = (
                    data.get(self.actor_id) or user_system.USER_SYSTEM_ID
                )
                data[self.actor_name] = (
                    data.get(self.actor_name) or user_system.USER_SYSTEM_NAME
                )
            actor_id = data.get(self.actor_id)
            actor_name = data.get(self.actor_name)
            if self.actor and actor_id is None:
                raise ValueError(
                    "Event {} requires an attribute specifying the actor_id".format(
                        self.event_type
                    )
                )
            if self.actor and (
                actor_id != user_system.USER_SYSTEM_ID and actor_name is None
            ):
                raise ValueError(
                    "Event {} requires an attribute specifying the actor_name".format(
                        self.event_type
                    )
                )
            if self.actor and (
                actor_id == user_system.USER_SYSTEM_ID and actor_name is None
            ):
                data[self.actor_name] = user_system.USER_SYSTEM_NAME

            if items:
                raise ValueError(
                    "Unknown attributes: {}".format(", ".join(items.keys()))
                )

            self.data = data

    @classmethod
    def get_event_subject(cls) -> Optional[str]:
        """Return the first part of the event_type

        e.g.

        >>> Event.event_type = 'experiment.deleted'
        >>> Event.get_event_subject() == 'experiment'
        """
        return event_context.get_event_subject(cls.event_type)

    @classmethod
    def get_event_action(cls) -> Optional[str]:
        """Return the second part of the event_type

        e.g.

        >>> Event.event_type = 'experiment.deleted'
        >>> Event.get_event_action() == 'deleted'
        """
        if not cls.actor:
            return None
        return event_context.get_event_action(cls.event_type)

    def serialize(
        self,
        dumps: bool = False,
        include_actor_name: bool = True,
        include_instance_info: bool = False,
    ) -> Union[str, Dict]:
        _data = self.data
        if not include_actor_name and self.actor and self.actor_name in _data:
            _data = copy.deepcopy(self.data)
            _data.pop(self.actor_name)
        data = {
            "uuid": self.uuid.hex,
            "timestamp": to_timestamp(self.datetime),
            "type": self.event_type,
            "ref_id": self.ref_id.hex if self.ref_id else None,
            "data": _data,
        }
        if include_instance_info:
            data["instance_id"] = self.instance_id
            data["instance_uuid"] = self.instance_uuid
        return dumps_htmlsafe(data) if dumps else data

    @classmethod
    def get_value_from_instance(cls, attr: str, instance: Any) -> Any:
        # Handle dot notation
        path = attr.split(".")
        value = instance
        for i in path:
            value = getattr(value, i, None)
            if value is None:
                break

        return value

    @staticmethod
    def get_instance_info(instance: Any) -> Dict:
        if isinstance(instance, Model):
            return {
                "instance_uuid": instance.uuid.hex
                if hasattr(instance, "uuid")
                else None,
                "instance_id": instance.id,
            }
        return {}

    @classmethod
    def from_instance(cls, instance: Any, **kwargs) -> "Event":
        values = {"instance": instance}
        if instance:
            values.update(cls.get_instance_info(instance))
        else:
            instance_id = kwargs.get("instance_id")
            if instance_id:
                values["instance_id"] = instance_id
            instance_uuid = kwargs.get("instance_uuid")
            if instance_uuid:
                values["instance_uuid"] = instance_uuid

        for attr in cls.get_event_attributes():
            # Convert dot notation
            value = kwargs.get(attr.name.replace(".", "_"))
            if value is None:
                value = cls.get_value_from_instance(attr=attr.name, instance=instance)
            values[attr.name] = value
        return cls(**values)

    @classmethod
    def from_event_data(cls, event_data: Mapping, **kwargs) -> "Event":
        return cls(
            datetime=to_datetime(event_data.get("timestamp")),
            uid=event_data.get("uuid"),
            ref_id=event_data.get("ref_id"),
            event_data=event_data.get("data"),
            instance=event_data.get("instance"),
            instance_id=event_data.get("instance_id"),
            instance_uuid=event_data.get("instance_uuid"),
            **kwargs,
        )


class ActorEvent(Event):
    actor = True


class ActorTopEntityEvent(ActorEvent):
    entity_uuid = "uuid"



import django.core.serializers.json

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0004_auto_20200905_1523"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifact",
            name="summary",
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name="artifactlineage",
            name="is_input",
            field=models.BooleanField(blank=True, default=False, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="inputs",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="meta_info",
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="outputs",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="params",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="status_conditions",
            field=models.JSONField(
                blank=True,
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="first_name",
            field=models.CharField(
                blank=True, max_length=150, verbose_name="first name"
            ),
        ),
    ]



from rest_framework.urlpatterns import format_suffix_patterns

from django.urls import re_path

from haupt.apis.projects import views
from haupt.common.apis.urls import projects

projects_urlpatterns = [
    re_path(projects.URLS_PROJECTS_CREATE, views.ProjectCreateView.as_view()),
    re_path(projects.URLS_PROJECTS_LIST, views.ProjectListView.as_view()),
    re_path(projects.URLS_PROJECTS_NAMES, views.ProjectNameListView.as_view()),
    re_path(projects.URLS_PROJECTS_DETAILS, views.ProjectDetailView.as_view()),
    re_path(projects.URLS_PROJECTS_ARCHIVE, views.ProjectArchiveView.as_view()),
    re_path(projects.URLS_PROJECTS_RESTORE, views.ProjectRestoreView.as_view()),
    re_path(projects.URLS_PROJECTS_BOOKMARK, views.ProjectBookmarkCreateView.as_view()),
    re_path(
        projects.URLS_PROJECTS_UNBOOKMARK, views.ProjectBookmarkDeleteView.as_view()
    ),
    re_path(projects.URLS_PROJECTS_STATS, views.ProjectStatsView.as_view()),
]

# Order is important, because the patterns could swallow other urls
urlpatterns = format_suffix_patterns(projects_urlpatterns)



from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

if settings.DB_ENGINE_NAME == "sqlite":
    ArrayField = None
else:
    from django.contrib.postgres.fields import ArrayField


class TagModel(models.Model):
    # Special handing for sqlite without migrating previous installation using ArrayField
    tags = (
        models.JSONField(
            encoder=DjangoJSONEncoder,
            blank=True,
            null=True,
        )
        if settings.DB_ENGINE_NAME == "sqlite"
        else ArrayField(
            base_field=models.CharField(max_length=64), blank=True, null=True
        )
    )

    class Meta:
        abstract = True



from clipped.utils.strings import strip_spaces
from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend
from rest_framework.filters import OrderingFilter as BaseOrderingFilter

from django.core.exceptions import ImproperlyConfigured
from django.db.models import F
from django.db.models.fields.json import KeyTransform

from haupt.common import query
from polyaxon.exceptions import PQLException


class QueryFilter(BaseFilterBackend):
    # The URL query parameter used for the ordering.
    query_param = "query"
    query_manager = None
    check_alive = False

    def get_check_alive(self, view):
        return getattr(view, "check_alive", self.check_alive)

    def get_query_param(self, view):
        return getattr(view, "query_param", self.query_param)

    def get_query_manager(self, view):
        query_manager = getattr(view, "query_manager", self.query_manager)
        if not query_manager:
            raise ImproperlyConfigured(
                "QueryFilter requires a query_manager to be set on the view."
            )
        return query_manager

    def process_alive(self, query_spec: str) -> str:
        if not query_spec:
            query_spec = "live_state:1"
        elif "live_state" not in query_spec:
            query_spec += ",live_state:1"

        return query_spec

    def filter_queryset(self, request, queryset, view):
        query_param = self.get_query_param(view=view)
        query_manager = self.get_query_manager(view=view)
        query_spec = request.query_params.get(query_param)
        check_alive = self.get_check_alive(view=view)
        if check_alive:
            query_spec = self.process_alive(query_spec=query_spec)

        if not query_spec and query_manager and query_manager.DEFAULT_FILTERS:
            query_spec = ",".join(
                "{}: {}".format(k, v) for k, v in query_manager.DEFAULT_FILTERS.items()
            )

        if query_spec and query_manager:
            try:
                queryset = query.filter_queryset(
                    manager=query_manager,
                    query_spec=query_spec,
                    queryset=queryset,
                    request=request,
                )
            except PQLException as e:
                if getattr(view, "exception_class", False):
                    raise e
                else:
                    raise ValidationError(e)

        return queryset


class OrderingFilter(BaseOrderingFilter):
    ordering_param = "sort"
    ordering_proxy_fields = {}

    def remove_invalid_fields(self, queryset, fields, view, request):
        result_fields = []
        annotation = {}
        proxy_fields = getattr(view, "ordering_proxy_fields", {}) or {}
        valid_fields = [
            item[0]
            for item in self.get_valid_fields(queryset, view, {"request": request})
        ]

        for field in fields:
            negation = "-" if field[0] == "-" else ""
            field = field.lstrip("-")
            if field in valid_fields:
                if negation:
                    result_fields.append(F(field).desc(nulls_last=True))
                else:
                    result_fields.append(field)
                continue

            field, suffix = query.parse_field(field)
            if field in proxy_fields:
                _proxy = proxy_fields[field]
                if _proxy.get("annotate"):
                    if negation:
                        result_fields.append(F(suffix).desc(nulls_last=True))
                    else:
                        result_fields.append(suffix)
                    annotation[suffix] = KeyTransform(suffix, _proxy["field"])
                else:
                    if negation:
                        result_fields.append(F(_proxy["field"]).desc(nulls_last=True))
                    else:
                        result_fields.append(_proxy["field"])

        return result_fields, annotation

    def get_ordering(self, request, queryset, view):
        """
        Ordering is set by a comma delimited ?ordering=... query parameter.

        The `ordering` query parameter can be overridden by setting
        the `ordering_param` value on the OrderingFilter or by
        specifying an `ORDERING_PARAM` value in the API settings.
        """
        params = request.query_params.get(self.ordering_param)
        if params:
            fields = strip_spaces(value=params, sep=",", join=False)
            ordering, annotations = self.remove_invalid_fields(
                queryset, fields, view, request
            )
            if ordering:
                return ordering, annotations

        # No ordering was included, or all the ordering fields were invalid
        return self.get_default_ordering(view), None

    def filter_queryset(self, request, queryset, view):
        ordering, annotations = self.get_ordering(request, queryset, view)

        if ordering:
            if annotations:
                queryset = queryset.annotate(**annotations)
            return queryset.order_by(*ordering)

        return queryset



from django.contrib.admin import site
from django.contrib.auth.admin import UserAdmin

from haupt.db.administration.artifacts import ArtifactAdmin, ArtifactLineageAdmin
from haupt.db.administration.projects import ProjectAdmin
from haupt.db.administration.runs import RunLightAdmin
from haupt.db.defs import Models

site.register(Models.User, UserAdmin)
site.register(Models.Artifact, ArtifactAdmin)
site.register(Models.ArtifactLineage, ArtifactLineageAdmin)
site.register(Models.Project, ProjectAdmin)
site.register(Models.Run, RunLightAdmin)



from typing import Dict, Optional

from haupt.schemas.platform_config import PlatformConfig


def set_celery(
    context, config: PlatformConfig, routes: Dict, schedules: Optional[Dict] = None
):
    context["CELERY_TASK_TRACK_STARTED"] = config.celery_task_track_started
    context["CELERY_BROKER_POOL_LIMIT"] = config.celery_broker_pool_limit
    context["CELERY_BROKER_BACKEND"] = config.broker_backend
    confirm_publish = config.celery_confirm_publish
    context["CELERY_CONFIRM_PUBLISH"] = confirm_publish
    if config.is_rabbitmq_broker and confirm_publish:
        # see https://github.com/celery/celery/issues/5410 for details
        context["CELERY_BROKER_TRANSPORT_OPTIONS"] = {"confirm_publish": True}

    context["CELERY_BROKER_URL"] = config.get_broker_url()

    context["INTERNAL_EXCHANGE"] = config.internal_exchange

    result_bucked = config.celery_result_backend
    if result_bucked:
        context["CELERY_RESULT_BACKEND"] = config.get_redis_url(
            config.celery_result_backend
        )

    context[
        "CELERY_WORKER_PREFETCH_MULTIPLIER"
    ] = config.celery_worker_prefetch_multiplier  # fmt: skip

    context["CELERY_TASK_ALWAYS_EAGER"] = config.celery_task_always_eager
    if config.celery_task_always_eager:
        context["CELERY_BROKER_TRANSPORT"] = "memory"

    context["CELERY_ACCEPT_CONTENT"] = ["application/json"]
    context["CELERY_TASK_DEFAULT_PRIORITY"] = 10
    context["CELERY_TASK_SERIALIZER"] = "json"
    context["CELERY_RESULT_SERIALIZER"] = "json"
    context["CELERY_TASK_IGNORE_RESULT"] = True
    context["CELERY_TIMEZONE"] = config.timezone
    context["CELERY_HARD_TIME_LIMIT_DELAY"] = config.celery_hard_time_limit_delay
    context[
        "CELERY_WORKER_MAX_TASKS_PER_CHILD"
    ] = config.celery_worker_max_tasks_per_child  # fmt: skip
    context[
        "CELERY_WORKER_MAX_MEMORY_PER_CHILD"
    ] = config.celery_worker_max_memory_per_child  # fmt: skip
    context["CELERY_TASK_ROUTES"] = routes
    if schedules:
        context["CELERY_BEAT_SCHEDULE"] = schedules

    context[
        "CLEANING_INTERVALS_ACTIVITY_LOGS"
    ] = config.cleaning_intervals_activity_logs  # fmt: skip
    context[
        "CLEANING_INTERVALS_NOTIFICATIONS"
    ] = config.cleaning_intervals_notifications  # fmt: skip
    context["CLEANING_INTERVALS_ARCHIVES"] = config.cleaning_intervals_archives
    context["CLEANING_INTERVALS_DELETION"] = config.cleaning_intervals_deletion



from haupt.apis.endpoints.run import RunArtifactEndpoint, RunResourceListEndpoint
from haupt.apis.methods import run_lineage as methods
from haupt.apis.serializers.artifacts import (
    RunArtifactNameSerializer,
    RunArtifactSerializer,
)
from haupt.apis.serializers.runs import (
    DownstreamRunEdgeSerializer,
    RunCloneSerializer,
    UpstreamRunEdgeSerializer,
)
from haupt.common.apis.filters import OrderingFilter, QueryFilter
from haupt.common.apis.paginator import LargeLimitOffsetPagination
from haupt.common.apis.regex import RUN_UUID_KEY
from haupt.common.endpoints.base import (
    CreateEndpoint,
    DestroyEndpoint,
    ListEndpoint,
    PostEndpoint,
    RetrieveEndpoint,
)
from haupt.common.events.registry.run import RUN_LINEAGE_ACTOR, RUN_NEW_ARTIFACTS
from haupt.db.defs import Models
from haupt.db.queries import artifacts as artifacts_queries
from haupt.db.query_managers.artifact import ArtifactQueryManager
from haupt.db.query_managers.run import RunQueryManager
from haupt.db.query_managers.run_edge import RunEdgeQueryManager


class RunArtifactListView(RunResourceListEndpoint, ListEndpoint, CreateEndpoint):
    queryset = artifacts_queries.artifacts
    serializer_class = RunArtifactSerializer
    pagination_class = LargeLimitOffsetPagination
    throttle_scope = "run-lineage"
    ALLOWED_METHODS = ["GET", "POST"]
    AUDITOR_EVENT_TYPES = {
        "POST": RUN_NEW_ARTIFACTS,
    }
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_EXTRA_KEYS = ("artifacts",)
    AUDIT_INSTANCE = True

    filter_backends = (QueryFilter, OrderingFilter)
    query_manager = ArtifactQueryManager
    check_alive = ArtifactQueryManager.CHECK_ALIVE
    ordering = ArtifactQueryManager.FIELDS_DEFAULT_ORDERING
    ordering_fields = ArtifactQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = ArtifactQueryManager.FIELDS_ORDERING_PROXY

    def create(self, request, *args, **kwargs):
        return methods.set_artifacts(view=self, request=request, *args, **kwargs)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["run"] = self.run
        return context


class RunArtifactNameListView(RunResourceListEndpoint, ListEndpoint):
    queryset = artifacts_queries.artifacts_names
    serializer_class = RunArtifactNameSerializer
    pagination_class = LargeLimitOffsetPagination
    ALLOWED_METHODS = ["GET"]


class RunEdgeListView(RunResourceListEndpoint, ListEndpoint):
    ALLOWED_METHODS = ["GET"]
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    filter_backends = (QueryFilter, OrderingFilter)
    query_manager = RunEdgeQueryManager
    check_alive = RunEdgeQueryManager.CHECK_ALIVE
    ordering_fields = RunEdgeQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = RunEdgeQueryManager.FIELDS_ORDERING_PROXY


class RunUpstreamListView(RunEdgeListView):
    queryset = Models.RunEdge.objects.prefetch_related("upstream").only(
        "values",
        "kind",
        "statuses",
        "upstream__id",
        "upstream__uuid",
        "upstream__name",
        "upstream__kind",
        "upstream__runtime",
    )
    serializer_class = UpstreamRunEdgeSerializer

    def enrich_queryset(self, queryset):
        return queryset.filter(downstream=self.run)


class RunDownstreamListView(RunEdgeListView):
    queryset = Models.RunEdge.objects.prefetch_related("downstream").only(
        "values",
        "kind",
        "statuses",
        "downstream__id",
        "downstream__uuid",
        "downstream__name",
        "downstream__kind",
        "downstream__runtime",
    )
    serializer_class = DownstreamRunEdgeSerializer

    def enrich_queryset(self, queryset):
        return queryset.filter(upstream=self.run)


class SetRunEdgesLineageView(RunResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": RUN_LINEAGE_ACTOR}
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_PROJECT_RESOURCES = True
    AUDIT_INSTANCE = True

    def create(self, request, *args, **kwargs):
        return methods.set_edges(view=self, request=request, *args, **kwargs)


class RunClonesListView(RunEdgeListView):
    queryset = Models.Run.objects.only(
        "id",
        "uuid",
        "name",
        "kind",
        "runtime",
        "cloning_kind",
        "status",
    )
    filter_backends = (QueryFilter, OrderingFilter)
    query_manager = RunQueryManager
    check_alive = RunQueryManager.CHECK_ALIVE
    ordering_fields = RunQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = RunQueryManager.FIELDS_ORDERING_PROXY
    serializer_class = RunCloneSerializer

    def enrich_queryset(self, queryset):
        return queryset.filter(original=self.run)


class RunArtifactDetailView(RunArtifactEndpoint, RetrieveEndpoint, DestroyEndpoint):
    queryset = artifacts_queries.artifacts
    serializer_class = RunArtifactSerializer
    ALLOWED_METHODS = ["GET", "DELETE"]

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["run"] = self.run_artifact.run
        return context



from haupt.common.events import event_actions, event_subjects
from haupt.common.events.event import ActorEvent, ActorTopEntityEvent
from haupt.common.events.registry.attributes import (
    PROJECT_OWNER_ATTRIBUTES,
    PROJECT_RESOURCE_OWNER_ATTRIBUTES,
)

PROJECT_ARCHIVED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.ARCHIVED, event_subjects.ACTOR
)
PROJECT_RESTORED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.RESTORED, event_subjects.ACTOR
)
RUN_ARCHIVED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.ARCHIVED, event_subjects.ACTOR
)
RUN_RESTORED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.RESTORED, event_subjects.ACTOR
)

EVENTS = {
    PROJECT_ARCHIVED_ACTOR,
    PROJECT_RESTORED_ACTOR,
    RUN_ARCHIVED_ACTOR,
    RUN_RESTORED_ACTOR,
}


class ProjectArchivedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_ARCHIVED_ACTOR
    actor = True
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectRestoredActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_RESTORED_ACTOR
    actor = True
    attributes = PROJECT_OWNER_ATTRIBUTES


class RunArchivedActorEvent(ActorEvent):
    event_type = RUN_ARCHIVED_ACTOR
    actor = True
    entity_uuid = "project.uuid"
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunRestoredActorEvent(ActorEvent):
    event_type = RUN_RESTORED_ACTOR
    actor = True
    entity_uuid = "project.uuid"
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES



from haupt.common import conf
from haupt.common.options.registry import cleaning

conf.subscribe(cleaning.CleaningIntervalsActivityLogs)
conf.subscribe(cleaning.CleaningIntervalsNotifications)
conf.subscribe(cleaning.CleaningIntervalsArchives)
conf.subscribe(cleaning.CleaningIntervalsDeletion)



from polyaxon._k8s.custom_resources import operation as op_crd
from polyaxon._k8s.manager.async_manager import AsyncK8sManager


async def get_k8s_operation(k8s_manager: AsyncK8sManager, resource_name: str):
    return await k8s_manager.get_custom_object(
        name=resource_name,
        group=op_crd.GROUP,
        version=op_crd.API_VERSION,
        plural=op_crd.PLURAL,
    )



import logging

from typing import Dict, List, Optional, Union

from clipped.utils.bools import to_bool
from clipped.utils.paths import delete_old_files
from rest_framework import status

from django.core.handlers.asgi import ASGIRequest
from django.db import transaction
from django.http import HttpResponse
from django.urls import path

from haupt.common.endpoints.files import FilePathResponse
from haupt.common.endpoints.validation import validate_internal_auth, validate_methods
from haupt.polyconf.config_manager import PLATFORM_CONFIG
from haupt.streams.connections.fs import AppFS
from haupt.streams.controllers.logs import get_archived_agent_logs
from haupt.streams.endpoints.base import UJSONResponse
from haupt.streams.tasks.op_spec import download_agent_spec, upload_agent_spec
from kubernetes_asyncio.client import V1Pod
from polyaxon import settings
from polyaxon._fs.async_manager import upload_data
from polyaxon._k8s.logging.async_monitor import (
    collect_agent_service_logs,
    get_agent_spec,
)
from polyaxon._k8s.manager.async_manager import AsyncK8sManager
from polyaxon._services import PolyaxonServices
from traceml.logging import V1Logs

_logger = logging.getLogger("haupt.streams.agents")


@transaction.non_atomic_requests
async def collect_agent_data(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    agent_uuid: str,
    methods: Optional[Dict] = None,
) -> Union[HttpResponse, FilePathResponse]:
    validate_methods(request, methods)
    try:
        validate_internal_auth(request)
    except Exception as e:
        errors = "Request requires an authenticated internal service %s" % e
        _logger.warning(errors)
        return UJSONResponse(
            data={"errors": errors},
            status=status.HTTP_400_BAD_REQUEST,
        )
    fs = await AppFS.get_fs()
    store_path = AppFS.get_fs_root_path()

    # Cleanup archives
    if store_path != settings.AGENT_CONFIG.local_root:
        delete_old_files(settings.AGENT_CONFIG.local_root)
    if store_path != settings.CLIENT_CONFIG.archives_root:
        delete_old_files(settings.CLIENT_CONFIG.archives_root)

    # Collect agent data
    async def collect_and_archive_agent_services_logs(pods: List[V1Pod]):
        for pod in pods:
            logs = await collect_agent_service_logs(k8s_manager=k8s_manager, pod=pod)

            for service in [
                PolyaxonServices.STREAMS.value,
                PolyaxonServices.OPERATOR.value,
                PolyaxonServices.AGENT.value,
            ]:
                if service in pod.metadata.name:
                    break
            if not service:
                continue

            last_file = 0
            for c_logs in V1Logs.chunk_logs(logs):
                last_file += 1
                subpath = ".agents/{}/logs/{}/{}".format(agent_uuid, service, last_file)
                await upload_data(
                    fs=fs, store_path=store_path, subpath=subpath, data=c_logs.to_json()
                )

    k8s_manager = AsyncK8sManager(
        namespace=namespace,
        in_cluster=settings.CLIENT_CONFIG.in_cluster,
    )
    await k8s_manager.setup()

    agent_spec, pods, _ = await get_agent_spec(k8s_manager=k8s_manager)
    if agent_spec:
        await upload_agent_spec(
            fs=fs,
            store_path=store_path,
            agent_uuid=agent_uuid,
            agent_spec=agent_spec,
        )
    # Do not collect core pods logs in debug mode
    if pods and PLATFORM_CONFIG.log_level != "DEBUG":
        await collect_and_archive_agent_services_logs(pods=pods)

    if k8s_manager:
        await k8s_manager.close()
    return UJSONResponse({})


@transaction.non_atomic_requests
async def k8s_inspect_agent(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    agent_uuid: str,
    methods: Optional[Dict] = None,
) -> Union[HttpResponse, FilePathResponse]:
    validate_methods(request, methods)
    connection = request.GET.get("connection")

    spec_path = await download_agent_spec(
        fs=await AppFS.get_fs(connection=connection),
        store_path=AppFS.get_fs_root_path(connection=connection),
        agent_uuid=agent_uuid,
    )
    if spec_path:
        return FilePathResponse(filepath=spec_path)
    return UJSONResponse({})


@transaction.non_atomic_requests
async def get_agent_logs(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    agent_uuid: str,
    methods: Optional[Dict] = None,
) -> UJSONResponse:
    validate_methods(request, methods)
    force = to_bool(request.GET.get("force"), handle_none=True)
    connection = request.GET.get("connection")
    service = request.GET.get("service")
    last_file = request.GET.get("last_file")

    operation_logs, last_file, files = await get_archived_agent_logs(
        fs=await AppFS.get_fs(connection=connection),
        store_path=AppFS.get_fs_root_path(connection=connection),
        agent_uuid=agent_uuid,
        service=service,
        last_file=last_file,
        check_cache=not force,
    )
    data = dict(
        last_time=None,
        last_file=last_file,
        logs=operation_logs,
        files=files,
    )
    return UJSONResponse(data)


URLS_AGENTS_COLLECT = "<str:namespace>/<str:owner>/agents/<str:agent_uuid>/collect"
URLS_AGENTS_K8S_INSPECT = (
    "<str:namespace>/<str:owner>/agents/<str:agent_uuid>/k8s_inspect"
)
URLS_AGENTS_LOGS = "<str:namespace>/<str:owner>/agents/<str:agent_uuid>/logs"


# fmt: off
internal_agent_routes = [
    path(
        URLS_AGENTS_COLLECT,
        collect_agent_data,
        name="collect_agent_data",
        kwargs=dict(methods=["POST"]),
    ),
]
agent_routes = [
    path(
        URLS_AGENTS_K8S_INSPECT,
        k8s_inspect_agent,
        name="k8s_inspect_agent",
        kwargs=dict(methods=["GET"]),
    ),
    path(
        URLS_AGENTS_LOGS,
        get_agent_logs,
        name="get_agent_logs",
        kwargs=dict(methods=["GET"]),
    ),
]



import os

from email.utils import formatdate
from typing import Dict, Optional

from clipped.utils.hashing import hash_value

from django.http import FileResponse


class FilePathResponse(FileResponse):
    def __init__(self, *args, as_attachment=False, filepath="", **kwargs):
        filename = os.path.basename(filepath) if filepath else ""
        headers = self.get_stat_headers(filepath)
        super().__init__(
            open(filepath, mode="rb"),
            *args,
            as_attachment=as_attachment,
            filename=filename,
            headers=headers,
            **kwargs,
        )

    def get_stat_headers(self, filepath: str) -> Optional[Dict]:
        if not filepath:
            return
        stat_result = os.stat(filepath)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = hash_value(etag_base.encode(), hash_length=None)

        return {"last-modified": last_modified, "etag": etag}



from haupt.proxies.schemas.base import get_config

OPTIONS = """
location = /favicon.ico {{
    rewrite ^ /static/images/favicon.ico;
}}
"""


def get_favicon_config():
    return get_config(options=OPTIONS, indent=0)



from django.core.validators import validate_slug
from django.db import models

from haupt.common.validation.blacklist import validate_blacklist_name


class NameableModel(models.Model):
    name = models.CharField(
        max_length=128,
        blank=True,
        null=True,
        default=None,
        validators=[validate_slug, validate_blacklist_name],
    )

    class Meta:
        abstract = True


class RequiredNameableModel(models.Model):
    name = models.CharField(
        max_length=128,
        validators=[validate_slug, validate_blacklist_name],
        db_index=True,
    )

    class Meta:
        abstract = True



from django.db import models


class ReadmeModel(models.Model):
    readme = models.TextField(blank=True, null=True)

    class Meta:
        abstract = True

    @property
    def has_readme(self) -> bool:
        return bool(self.readme)



from typing import Dict, Optional

from clipped.compact.pydantic import ValidationError as PydanticValidationError
from rest_framework import fields, serializers
from rest_framework.exceptions import ValidationError

from django.db import IntegrityError

from haupt.apis.serializers.base.bookmarks_mixin import BookmarkedSerializerMixin
from haupt.apis.serializers.base.cloning import CloningMixin
from haupt.apis.serializers.base.is_managed import IsManagedMixin
from haupt.apis.serializers.base.owner_mixin import ProjectOwnerMixin
from haupt.apis.serializers.base.pipeline import PipelineMixin
from haupt.apis.serializers.base.project import ProjectMixin
from haupt.apis.serializers.base.resources_mixin import ResourcesMixin
from haupt.apis.serializers.base.settings import FullSettingsMixin, SettingsMixin
from haupt.apis.serializers.base.tags import TagsMixin
from haupt.apis.serializers.base.user_mixin import UserMixin
from haupt.db.defs import Models
from haupt.db.managers.runs import create_run
from haupt.orchestration import operations
from polyaxon._polyaxonfile import OperationSpecification
from polyaxon.exceptions import PolyaxonException
from polyaxon.schemas import ManagedBy, V1RunEdgeKind, V1RunPending


class BaseRunEdgeSerializer(serializers.ModelSerializer):
    class Meta:
        model = Models.RunEdge
        fields = (
            "values",
            "kind",
            "statuses",
        )

    def get_run(self, obj: Models.Run) -> Optional[Dict]:
        return {
            "uuid": obj.uuid.hex,
            "name": obj.name,
            "kind": obj.kind,
            "runtime": obj.runtime,
        }


class UpstreamRunEdgeSerializer(BaseRunEdgeSerializer):
    upstream = fields.SerializerMethodField()

    class Meta(BaseRunEdgeSerializer.Meta):
        fields = BaseRunEdgeSerializer.Meta.fields + ("upstream",)

    def get_upstream(self, obj: Models.RunEdge):
        if not obj.upstream_id:
            return None
        return self.get_run(obj.upstream)


class DownstreamRunEdgeSerializer(BaseRunEdgeSerializer):
    downstream = fields.SerializerMethodField()

    class Meta(BaseRunEdgeSerializer.Meta):
        fields = BaseRunEdgeSerializer.Meta.fields + ("downstream",)

    def get_downstream(self, obj: Models.RunEdge):
        if not obj.downstream_id:
            return None
        return self.get_run(obj.downstream)


class RunStatusSerializer(serializers.ModelSerializer):
    uuid = fields.UUIDField(format="hex", read_only=True)
    condition = fields.DictField(
        write_only=True, allow_empty=True, allow_null=True, required=False
    )
    force = fields.BooleanField(write_only=True, required=False)

    class Meta:
        model = Models.Run
        fields = (
            "uuid",
            "status",
            "condition",
            "status_conditions",
            "meta_info",
            "force",
        )
        extra_kwargs = {
            "status": {"read_only": True},
            "status_conditions": {"read_only": True},
        }


class RunCloneSerializer(serializers.ModelSerializer):
    class Meta:
        model = Models.Run
        fields = (
            "uuid",
            "name",
            "kind",
            "status",
            "runtime",
            "cloning_kind",
        )


class BaseRunSerializer(
    serializers.ModelSerializer,
    CloningMixin,
    PipelineMixin,
    SettingsMixin,
    TagsMixin,
):
    uuid = fields.UUIDField(format="hex", read_only=True)
    original = fields.SerializerMethodField()
    pipeline = fields.SerializerMethodField()
    started_at = fields.DateTimeField(read_only=True)
    finished_at = fields.DateTimeField(read_only=True)
    schedule_at = fields.DateTimeField(read_only=True)
    settings = fields.SerializerMethodField()

    class Meta:
        model = Models.Run
        fields = (
            "uuid",
            "name",
            "created_at",
            "updated_at",
            "started_at",
            "finished_at",
            "schedule_at",
            "wait_time",
            "duration",
            "kind",
            "runtime",
            "meta_info",
            "status",
            "pipeline",
            "original",
            "is_managed",
            "managed_by",
            "pending",
            "inputs",
            "outputs",
            "tags",
            "settings",
        )
        extra_kwargs = {
            "is_managed": {"read_only": True},
            "managed_by": {"read_only": True},
        }


class RunSerializer(
    BaseRunSerializer,
    UserMixin,
    ProjectMixin,
    ProjectOwnerMixin,
):
    owner = fields.SerializerMethodField()
    user = fields.SerializerMethodField()
    project = fields.SerializerMethodField()

    class Meta(BaseRunSerializer.Meta):
        fields = BaseRunSerializer.Meta.fields + (
            "user",
            "owner",
            "project",
        )


class BookmarkedRunSerializer(
    BaseRunSerializer, UserMixin, ResourcesMixin, BookmarkedSerializerMixin
):
    bookmarked_model = "run"

    resources = fields.SerializerMethodField()
    user = fields.SerializerMethodField()

    class Meta(BaseRunSerializer.Meta):
        fields = BaseRunSerializer.Meta.fields + (
            "user",
            "bookmarked",
            "resources",
        )


class BookmarkedTimelineRunSerializer(BookmarkedRunSerializer):
    class Meta:
        model = Models.Run
        fields = (
            "uuid",
            "name",
            "created_at",
            "updated_at",
            "started_at",
            "finished_at",
            "schedule_at",
            "wait_time",
            "duration",
            "kind",
            "runtime",
            "status",
            "tags",
            "user",
            "bookmarked",
            "pipeline",
            "original",
            "is_managed",
            "managed_by",
            "pending",
            "tags",
        )


class GraphRunSerializer(RunSerializer, UserMixin):
    graph = fields.SerializerMethodField()

    class Meta:
        model = Models.Run
        fields = (
            "uuid",
            "name",
            "created_at",
            "updated_at",
            "started_at",
            "finished_at",
            "schedule_at",
            "wait_time",
            "duration",
            "kind",
            "runtime",
            "status",
            "tags",
            "user",
            "pipeline",
            "pending",
            "graph",
        )

    def get_graph(self, obj):
        graph = self.context.get("graph", None)

        if graph is not None:
            return graph.get(obj.uuid.hex)
        return None


class RunDetailSerializer(
    RunSerializer,
    FullSettingsMixin,
    IsManagedMixin,
    ResourcesMixin,
    BookmarkedSerializerMixin,
):
    bookmarked_model = "run"
    merge = fields.BooleanField(write_only=True, required=False)
    resources = fields.SerializerMethodField()

    class Meta(RunSerializer.Meta):
        fields = RunSerializer.Meta.fields + (
            "readme",
            "description",
            "raw_content",
            "content",
            "live_state",
            "merge",
            "bookmarked",
            "resources",
        )
        extra_kwargs = {
            "content": {"read_only": True},
            "raw_content": {"read_only": True},
        }

    def validate(self, attrs):
        attrs = super().validate(attrs)
        attrs = self.check_if_entity_is_managed(attrs=attrs, entity_name="Run")
        return attrs

    def validated_inputs(self, validated_data, inputs):
        new_inputs = validated_data.get("inputs")
        if not validated_data.get("merge") or not inputs or not new_inputs:
            # This is the default behavior
            return validated_data

        inputs.update(new_inputs)
        validated_data["inputs"] = inputs
        return validated_data

    def validated_meta(self, validated_data, meta_info):
        new_meta_info = validated_data.get("meta_info")
        if not validated_data.get("merge") or not meta_info or not new_meta_info:
            # This is the default behavior
            return validated_data

        meta_info.update(new_meta_info)
        validated_data["meta_info"] = meta_info
        return validated_data

    def validated_outputs(self, validated_data, outputs):
        new_outputs = validated_data.get("outputs")
        if not validated_data.get("merge") or not outputs or not new_outputs:
            # This is the default behavior
            return validated_data

        outputs.update(new_outputs)
        validated_data["outputs"] = outputs
        return validated_data

    def update(self, instance, validated_data):
        validated_data = self.validated_tags(
            validated_data=validated_data, tags=instance.tags
        )
        validated_data = self.validated_outputs(
            validated_data=validated_data, outputs=instance.outputs
        )
        validated_data = self.validated_inputs(
            validated_data=validated_data, inputs=instance.inputs
        )
        validated_data = self.validated_meta(
            validated_data=validated_data, meta_info=instance.meta_info
        )

        return super().update(instance=instance, validated_data=validated_data)


class OfflineRunSerializer(
    serializers.ModelSerializer,
    IsManagedMixin,
    TagsMixin,
):
    uuid = fields.UUIDField(format="hex")
    created_at = fields.DateTimeField()

    class Meta:
        model = Models.Run
        fields = (
            "uuid",
            "name",
            "description",
            "tags",
            "created_at",
            "updated_at",
            "started_at",
            "finished_at",
            "wait_time",
            "duration",
            "kind",
            "runtime",
            "meta_info",
            "status",
            "status_conditions",
            "is_managed",
            "managed_by",
            "inputs",
            "outputs",
            "content",
            "raw_content",
        )

    def validate(self, attrs):
        attrs = super().validate(attrs)
        attrs = self.check_if_entity_is_managed(attrs=attrs, entity_name="Run")
        return attrs

    def create(self, validated_data, commit: bool = True):
        try:
            obj = super().create(validated_data)
        except IntegrityError:
            raise ValidationError(
                f"A run with uuid {validated_data.get('uuid')} already exists."
            )
        # Override auto-field for created_at
        created_at = validated_data.get("created_at")
        if created_at:
            obj.created_at = created_at
        if commit:
            obj.save()
        return obj


class OperationCreateSerializer(serializers.ModelSerializer, IsManagedMixin, TagsMixin):
    uuid = fields.UUIDField(format="hex", read_only=True)
    is_approved = fields.BooleanField(write_only=True, allow_null=True, required=False)
    settings = fields.SerializerMethodField()

    class Meta:
        model = Models.Run
        fields = (
            "uuid",
            "kind",
            "name",
            "description",
            "content",
            "is_managed",
            "managed_by",
            "pending",
            "meta_info",
            "tags",
            "is_approved",
            "settings",
        )
        extra_kwargs = {
            "is_approved": {"write_only": True},
            "kind": {"read_only": True},
        }

    def get_settings(self, obj):
        settings = {}
        build = (
            obj.upstream_runs.filter(downstream_edges__kind=V1RunEdgeKind.BUILD)
            .order_by("created_at")
            .last()
        )
        if build:
            settings["build"] = {
                "name": build.name,
                "status": build.status,
                "uuid": build.uuid.hex,
            }

        agent = None
        if hasattr(obj, "agent") and obj.agent:
            agent = {
                "name": obj.agent.name,
                "version": obj.agent.version,
                "url": obj.agent.hostname,
            }
        if agent:
            settings["agent"] = agent
        return settings

    def validate(self, attrs):
        attrs = super().validate(attrs)
        attrs = self.check_if_entity_is_managed(attrs=attrs, entity_name="Run")
        return attrs

    def create(self, validated_data):
        managed_by = validated_data.get("managed_by")
        if managed_by is None:
            managed_by = ManagedBy.AGENT
        content = validated_data.get("content")
        meta_info = validated_data.get("meta_info") or {}

        if managed_by is None:
            raise ValidationError("Run is not validated correctly")

        is_managed = ManagedBy.is_managed(managed_by)
        if is_managed and not content:
            raise ValidationError(
                "Managed runs require a content with valid specification"
            )

        project_id = validated_data["project"].id
        user = validated_data.get("user")
        name = validated_data.get("name")
        description = validated_data.get("description")
        tags = validated_data.get("tags")
        pending = validated_data.get("pending")
        # Check the deprecated `is_approved` flag
        if pending is None:
            is_approved = validated_data.get("is_approved")
            if is_approved is False:
                pending = V1RunPending.UPLOAD

        if is_managed or content:
            try:
                op_spec = OperationSpecification.read(content)
            except Exception as e:
                raise ValidationError(e)
            if op_spec.is_template():
                raise ValidationError(
                    "Received a template polyaxonfile, "
                    "Please customize the specification or disable the template."
                )
            try:
                return operations.init_and_save_run(
                    project_id=project_id,
                    user_id=user.id if user else None,
                    op_spec=op_spec,
                    name=name,
                    description=description,
                    tags=tags,
                    meta_info=meta_info,
                    managed_by=managed_by,
                    pending=pending,
                    supported_kinds=validated_data.get("supported_kinds"),
                    supported_owners=validated_data.get("supported_owners"),
                )
            except (PydanticValidationError, PolyaxonException, ValueError) as e:
                raise ValidationError(e)
        else:
            return create_run(
                project_id=project_id,
                user_id=user.id if user else None,
                name=name,
                description=description,
                tags=tags,
                meta_info=meta_info,
                managed_by=managed_by,
            )



from celery import Celery, states

from django.apps import apps

STATES = states

app = Celery("haupt")

app.config_from_object("django.conf:settings", namespace="CELERY")
app.autodiscover_tasks(lambda: [n.name for n in apps.get_app_configs()])



from clipped.utils.bools import to_bool
from rest_framework.pagination import LimitOffsetPagination


class PolyaxonBasePagination(LimitOffsetPagination):
    no_page_query_param = "no_page"

    def get_no_page(self, request):
        try:
            return to_bool(
                request.query_params.get(self.no_page_query_param),
                handle_none=True,
            )
        except (KeyError, ValueError):
            return False

    def get_count(self, queryset):
        if self.no_page:
            return None
        return super().get_count(queryset)

    def paginate_queryset(self, queryset, request, view=None):
        self.no_page = self.get_no_page(request)
        self.limit = self.get_limit(request)
        if self.limit is None:
            return None

        self.count = self.get_count(queryset)
        self.offset = self.get_offset(request)
        self.request = request

        if self.no_page:
            return queryset[self.offset : self.offset + self.limit]

        if self.count > self.limit and self.template is not None:
            self.display_page_controls = True

        if self.count == 0 or self.offset > self.count:
            return []
        return queryset[self.offset : self.offset + self.limit]

    def get_next_link(self):
        if self.no_page:
            return None
        return super().get_next_link()

    def get_previous_link(self):
        if self.no_page:
            return None
        return super().get_previous_link()


class LargeLimitOffsetPagination(PolyaxonBasePagination):
    max_limit = None
    default_limit = 100


class PolyaxonPagination(PolyaxonBasePagination):
    max_limit = 100
    default_limit = 20



from haupt.db.administration.utils import DiffModelAdmin


class ArtifactAdmin(DiffModelAdmin):
    list_display = (
        "name",
        "kind",
        "state",
    )
    list_display_links = (
        "name",
        "kind",
        "state",
    )
    readonly_fields = DiffModelAdmin.readonly_fields + ("name",)
    fields = ("name", "kind", "created_at", "updated_at")


class ArtifactLineageAdmin(DiffModelAdmin):
    list_display = (
        "artifact",
        "run",
        "is_input",
    )
    list_display_links = (
        "artifact",
        "run",
        "is_input",
    )
    readonly_fields = DiffModelAdmin.readonly_fields + ("artifact", "run")
    fields = ("artifact", "run", "is_input", "created_at", "updated_at")



from django.conf import settings
from django.db import models


class ContributorsModel(models.Model):
    contributors = models.ManyToManyField(
        settings.AUTH_USER_MODEL, blank=True, related_name="+"
    )

    class Meta:
        abstract = True



from haupt import settings
from haupt.common import conf
from haupt.common.options.registry.installation import (
    ORGANIZATION_ID,
    PLATFORM_DIST,
    PLATFORM_MODE,
    PLATFORM_VERSION,
)


def get_version():
    data = {
        "key": conf.get(ORGANIZATION_ID),
        "version": conf.get(PLATFORM_VERSION),
        "dist": conf.get(PLATFORM_DIST),
        "mode": conf.get(PLATFORM_MODE),
    }
    if settings.SANDBOX_CONFIG:
        data["mode"] = settings.SANDBOX_CONFIG.mode

    return data



from clipped.compact.pydantic import ValidationError as PydanticValidationError
from clipped.utils.lists import to_list
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from django.db.models import Q

from haupt.db.defs import Models
from polyaxon._flow import V1RunEdgeKind
from polyaxon._sdk.schemas import V1RunEdgesGraph
from traceml.artifacts import V1RunArtifact, V1RunArtifacts


def set_artifacts(view, request, *args, **kwargs):
    if not request.data:
        raise ValidationError("Received no artifacts.")

    data = request.data
    if "artifacts" in data:
        data = data["artifacts"]
        try:
            V1RunArtifacts.from_dict(request.data)
        except PydanticValidationError as e:
            raise ValidationError(e)
    else:  # Backward compatibility with legacy clients
        data = to_list(data)
        try:
            [V1RunArtifact.from_dict(r) for r in data]
        except PydanticValidationError as e:
            raise ValidationError(e)

    view.audit(request, *args, **kwargs, artifacts=data)
    return Response(status=status.HTTP_201_CREATED)


def set_edges(view, request, *args, **kwargs):
    if not request.data:
        raise ValidationError("Received no edges.")

    data = request.data
    if "edges" not in data:
        return Response(status=status.HTTP_400_BAD_REQUEST)

    try:
        lineage = V1RunEdgesGraph.from_dict(data)
    except PydanticValidationError as e:
        raise ValidationError(e)

    Models.RunEdge.objects.filter(
        Q(downstream=view.run) | Q(upstream=view.run)
    ).delete()
    if not lineage.edges:
        return Response(status=status.HTTP_201_CREATED)

    runs = Models.Run.all.filter(uuid__in=[e.run for e in lineage.edges]).values(
        "uuid", "id"
    )
    runs_uuid_to_id = {r["uuid"].hex: r["id"] for r in runs}
    edges = []
    for e in lineage.edges:
        if e.is_upstream:
            upstream = runs_uuid_to_id.get(e.run)
            downstream = view.run.id
        else:
            upstream = view.run.id
            downstream = runs_uuid_to_id.get(e.run)
        edges.append(
            Models.RunEdge(
                upstream_id=upstream,
                downstream_id=downstream,
                values=e.values,
                kind=V1RunEdgeKind.MANUAL,
            )
        )
    Models.RunEdge.objects.bulk_create(edges)
    view.audit(request, *args, **kwargs, artifacts=data)
    return Response(status=status.HTTP_201_CREATED)



from collections import namedtuple
from typing import Any, Optional

from clipped.utils.http import absolute_uri

from haupt.common import user_system
from haupt.common.events import event_subjects
from polyaxon._utils.urls_utils import get_fqn_run_url, get_owner_url, get_project_url


class EventItemContextSpec(namedtuple("EventItemContextSpec", "name url object_id")):
    pass


class EventContextSpec(
    namedtuple(
        "EventItemContextSpec",
        "subject_action subject action actor_context object_context datetime",
    )
):
    pass


def get_event_subject(event_type: str) -> str:
    """Return the first part of the event_type

    e.g.

    >>> Event.event_type = 'experiment.deleted'
    >>> Event.get_event_subject() == 'experiment'
    """
    return event_type.split(".")[0]


def get_event_action(event_type: str) -> str:
    """Return the second part of the event_type

    e.g.

    >>> Event.event_type = 'experiment.deleted'
    >>> Event.get_event_action() == 'deleted'
    """
    return event_type.split(".")[1]


def get_event_actor_context(
    event: "Event",  # noqa: F821
) -> Optional[EventItemContextSpec]:
    if not event.actor:
        return None

    username = event.data.get(event.actor_name)
    if username is None:
        return None
    if username == user_system.USER_SYSTEM_NAME:
        return EventItemContextSpec(name=username, url="/", object_id=None)
    return EventItemContextSpec(
        name=username, url=get_owner_url(username), object_id=None
    )


def get_event_object_context(
    event_content_object: Any, event_type: str
) -> Optional[EventItemContextSpec]:
    # Deleted objects don't have a content object any more
    if not event_content_object:
        return EventItemContextSpec(name=None, url=None, object_id=None)

    event_subject = get_event_subject(event_type)

    object_id = None
    object_url = None
    object_name = None
    if hasattr(event_content_object, "id"):
        object_id = event_content_object.id

    if hasattr(event_content_object, "unique_name"):
        object_name = event_content_object.unique_name
        if event_subject == event_subjects.PROJECT:
            object_url = get_project_url(object_name)
        elif event_subject == event_subjects.RUN:
            object_url = get_fqn_run_url(object_name)

    elif hasattr(event_content_object, "name"):
        object_name = event_content_object.name
    elif hasattr(event_content_object, "username"):
        object_name = event_content_object.username
        object_url = get_owner_url(event_content_object.username)

    # Set proper url
    object_url = absolute_uri("ui{}".format(object_url))
    return EventItemContextSpec(name=object_name, url=object_url, object_id=object_id)


def get_event_context(event: "Event") -> EventContextSpec:  # noqa: F821
    subject = get_event_subject(event_type=event.event_type)
    action = get_event_action(event_type=event.event_type)
    actor_context = get_event_actor_context(event=event)
    object_context = get_event_object_context(
        event_content_object=event.instance, event_type=event.event_type
    )
    return EventContextSpec(
        subject_action="{} {}".format(subject, action),
        subject=subject,
        action=action,
        actor_context=actor_context,
        object_context=object_context,
        datetime=event.datetime,
    )


def get_readable_event(event_context: EventContextSpec) -> str:
    description = "{} on {}".format(
        event_context.subject_action, event_context.datetime
    )
    if event_context.actor_context:
        description += "\nActor: [{}]({})".format(
            event_context.actor_context.name, event_context.actor_context.url
        )

    if event_context.object_context.name and event_context.object_context.url:
        description += "\nObject: [{}]({})".format(
            event_context.object_context.name, event_context.object_context.url
        )

    return description



from haupt.db.query_managers.manager import BaseQueryManager
from polyaxon._pql.builder import BoolCondition, SearchCondition, ValueCondition
from polyaxon._pql.parser import parse_search_operation, parse_value_operation


class ArtifactQueryManager(BaseQueryManager):
    NAME = "artifact"
    FIELDS_ORDERING = ("name", "kind", "path", "is_input")
    FIELDS_USE_UUID = {"run"}
    FIELDS_PROXY = {
        "id": "name",
        "name": "artifact__name",
        "kind": "artifact__kind",
        "path": "artifact__path",
        "state": "artifact__state",
    }
    CHECK_ALIVE = False
    PARSERS_BY_FIELD = {
        # Name
        "name": parse_search_operation,
        # Kind
        "kind": parse_value_operation,
        # Path
        "path": parse_value_operation,
        # State
        "state": parse_value_operation,
        # Is input
        "is_input": parse_value_operation,
        # Run
        "run": parse_value_operation,
    }
    CONDITIONS_BY_FIELD = {
        # Name
        "name": SearchCondition,
        # Kind
        "kind": ValueCondition,
        # Path
        "path": ValueCondition,
        # State
        "state": ValueCondition,
        # Is input
        "is_input": BoolCondition,
        # Run
        "run": ValueCondition,
    }



REST_FRAMEWORK = {
    "DEFAULT_RENDERER_CLASSES": ("rest_framework.renderers.JSONRenderer",),
    "DEFAULT_THROTTLE_RATES": {"health": "{}/min".format(10)},
    "DEFAULT_VERSIONING_CLASS": "rest_framework.versioning.NamespaceVersioning",
    "DEFAULT_PAGINATION_CLASS": "haupt.common.apis.paginator.PolyaxonPagination",
}



from clipped.decorators.memoization import memoize

from django.db import connection


class RawBulkInserter:
    """
    This class does bulk inserts similar to Django's bulk_create, but does not use model instances.
    This can save a significant amount of time for very large inserts (e.g. > 100k values)
    IMPORTANT: Since all of the ORM magic is bypassed, you have to be careful with the types of
    the values you pass.
    """

    def __init__(
        self, model, fields, chunk_size=40000, fetch_ids=False, table_name=None
    ):
        """
        Params:
            model: model class
            fields: iterable of field names to be used for inserts
                For foreign keys, use <field_name>_id, not just <fieldname>
            chunk_size: maximum number of rows for a single insert
            fetch_ids: if True, add_row and insert will return a list of created ids
                Requires that the model has an 'id' column
        """
        self._chunk_size = chunk_size
        self._fetch_ids = fetch_ids
        if fetch_ids:
            self._created_ids = []

        self._num_fields = len(fields)

        if model is not None:
            table_name = model._meta.db_table

        # create a memoized method to generate sql based on the number of rows to be inserted
        self._generate_sql = memoize(
            self._create_sql_generator(table_name, fields, fetch_ids)
        )

        self._num_queued_rows = 0

        self._values = []
        self._cursor = connection.cursor()

    def queue_row(self, *values):
        """
        Add a row and bulk insert if internal queue is larger than CHUNK_SIZE.
        WARNING: Most of the usual Django type conversion magic is bypassed, so the inserted data
        MIGHT look different for anything but simple number/string fields.
        For date/time, always use (timezone) "aware" objects (e.g. datetime.now(tz=utc) with utc
        from django.utils.timezone)
        """
        if len(values) != self._num_fields:
            raise ValueError("Not enough values, expected {}".format(self._num_fields))

        self._values.extend(values)
        self._num_queued_rows += 1

        if self._num_queued_rows >= self._chunk_size:
            self.insert()

    def insert(self):
        """
        Triggers the insertion of previously added rows. This should be done once after all rows
        have been queued.
        """
        if self._num_queued_rows == 0:
            return

        self._cursor.execute(self._generate_sql(self._num_queued_rows), self._values)

        if self._fetch_ids:
            rows = self._cursor.fetchall()
            self._created_ids.extend([row[0] for row in rows])

        self._values = []
        self._num_queued_rows = 0

    @property
    def all_created_ids(self):
        """
        Id's of all created rows, in the same order as the rows were queued (-> queue_row)
        Requires fetch_ids=True in the constructor.
        IMPORTANT: May be empty/incomplete before full insert is triggered by insert()
        """
        return self._created_ids

    @staticmethod
    def _create_sql_generator(table_name, fields, fetch_ids):
        # escape field names that contain a '%'
        fields = [field.replace("%", "%%") for field in fields]

        base_sql = 'INSERT INTO {} ("{}") VALUES '.format(
            table_name, '","'.join(fields)
        )
        placeholder_single_row = "({})".format((",".join(["%s"] * len(fields))))

        def generator(num_rows):
            sql_list = [base_sql, ",".join([placeholder_single_row] * num_rows)]
            if fetch_ids:
                sql_list.append(" RETURNING id")

            return "".join(sql_list)

        return generator



from clipped.decorators.memoization import memoize

from django.conf import settings
from django.db import connection


class RawBulkUpdater:
    """
    This class does bulk updates and behaves the same way as the RawBulkInserter
    """

    def __init__(self, model, query_fields, value_fields, chunk_size=40000):
        """
        Params:
            model: model class
            query_fields: iterable of query field names to be used for the where clause
                       For foreign keys, use <field_name>_id, not just <fieldname>
            value_fields: iterable of value field names to be used for the updating
                       For foreign keys, use <field_name>_id, not just <fieldname>
            chunk_size: maximum number of rows for a single insert
        """
        if settings.DB_ENGINE_NAME == "sqlite":
            raise ValueError("Sqlite does not support bulk updates")
        self._chunk_size = chunk_size

        self._num_query_values = len(query_fields)
        self._num_values = len(value_fields)

        # create a memoized method to generate sql based on the number of rows to be inserted
        self._generate_sql = memoize(
            self._update_sql_generator(model, query_fields, value_fields)
        )

        self._num_queued_rows = 0

        self._values = []
        self._cursor = connection.cursor()

    def queue_row(self, queries, values):
        """
        Add a row and bulk update if internal queue is larger than CHUNK_SIZE.
        WARNING: Most of the usual Django type conversion magic is bypassed, so the inserted data
        MIGHT look different for anything but simple number/string fields.
        For date/time, always use (timezone) "aware" objects (e.g. datetime.now(tz=utc) with utc
        from django.utils.timezone)
        """
        if len(queries) != self._num_query_values:
            raise ValueError(
                "Not enough keys, expected {}".format(self._num_query_values)
            )

        if len(values) != self._num_values:
            raise ValueError("Not enough values, expected {}".format(self._num_values))

        self._values.extend(queries + values)
        self._num_queued_rows += 1

        if self._num_queued_rows >= self._chunk_size:
            self.update()

    def update(self):
        """
        Triggers the update of previously added rows. This should be done once after all rows
        have been queued.
        """
        if self._num_queued_rows == 0:
            return

        sql = self._generate_sql(self._num_queued_rows)
        self._cursor.execute(sql, self._values)

        self._values = []
        self._num_queued_rows = 0

    def _update_sql_generator(self, model, query_fields, value_fields):
        db_table = model._meta.db_table
        fields = {f.name: f for f in model._meta.fields}
        base_sql = "UPDATE {} SET {} FROM ".format(
            db_table,
            ",".join(
                [
                    "{}=tempvalues.{}".format(value_field, value_field)
                    for value_field in value_fields
                ]
            ),
        )
        # In psycopg2 this was:
        # placeholder_single_row = "({})".format(
        #     (",".join(["%s"] * (len(query_fields) + len(value_fields))))
        # )
        # After the upgrade we have to mimic the update logic in the SQLUpdateCompiler.as_sql
        update_fields = ["%s"] * len(query_fields)  # Retrieve fields
        update_fields += [
            fields[vf].get_placeholder(None, None, connection)
            if hasattr(fields[vf], "get_placeholder")
            else "%s"
            for vf in value_fields
        ]  # Update fields after casting to correct type
        placeholder_single_row = "({})".format(",".join(update_fields))
        single_where_statement = ".".join(["{}".format(db_table), "{}=tempvalues.{}"])
        where_statement_sql = "AS tempvalues ({},{}) WHERE {}".format(
            ",".join([query_field for query_field in query_fields]),
            ",".join([value_field for value_field in value_fields]),
            " AND ".join(
                [
                    single_where_statement.format(query_field, query_field)
                    for query_field in query_fields
                ]
            ),
        )

        def generator(num_rows):
            placeholder_all_rows = "(VALUES {}) ".format(
                ",".join([placeholder_single_row] * num_rows)
            )
            sql_list = [base_sql, placeholder_all_rows, where_statement_sql]

            return "".join(sql_list)

        return generator



from haupt.common.events import event_actions, event_subjects
from haupt.common.events.event import ActorEvent, Attribute

ARTIFACT_VERSION_DELETED = "{}.{}".format(
    event_subjects.ARTIFACT_VERSION, event_actions.DELETED
)
ARTIFACT_VERSION_CREATED_ACTOR = "{}.{}.{}".format(
    event_subjects.ARTIFACT_VERSION, event_actions.CREATED, event_subjects.ACTOR
)
ARTIFACT_VERSION_UPDATED_ACTOR = "{}.{}.{}".format(
    event_subjects.ARTIFACT_VERSION, event_actions.UPDATED, event_subjects.ACTOR
)
ARTIFACT_VERSION_DELETED_ACTOR = "{}.{}.{}".format(
    event_subjects.ARTIFACT_VERSION, event_actions.DELETED, event_subjects.ACTOR
)
ARTIFACT_VERSION_VIEWED_ACTOR = "{}.{}.{}".format(
    event_subjects.ARTIFACT_VERSION, event_actions.VIEWED, event_subjects.ACTOR
)
ARTIFACT_VERSION_TRANSFERRED_ACTOR = "{}.{}.{}".format(
    event_subjects.ARTIFACT_VERSION, event_actions.TRANSFERRED, event_subjects.ACTOR
)
ARTIFACT_VERSION_NEW_STAGE = "{}.{}".format(
    event_subjects.ARTIFACT_VERSION, event_actions.NEW_STAGE
)

EVENTS = {
    ARTIFACT_VERSION_DELETED,
    ARTIFACT_VERSION_CREATED_ACTOR,
    ARTIFACT_VERSION_UPDATED_ACTOR,
    ARTIFACT_VERSION_VIEWED_ACTOR,
    ARTIFACT_VERSION_DELETED_ACTOR,
    ARTIFACT_VERSION_TRANSFERRED_ACTOR,
    ARTIFACT_VERSION_NEW_STAGE,
}

ARTIFACT_VERSION_ATTRIBUTES = (
    Attribute("uuid", is_uuid=True),
    Attribute("name", is_required=False),
    Attribute("owner_id"),
    Attribute("project.uuid", is_uuid=True),
    Attribute("project.name"),
    Attribute("project.owner.name"),
)


class VersionActorEvent(ActorEvent):
    entity_uuid = "project.uuid"


class ArtifactVersionDeletedEvent(VersionActorEvent):
    event_type = ARTIFACT_VERSION_DELETED
    attributes = (Attribute("id"),)


class ArtifactVersionCreatedActorEvent(VersionActorEvent):
    event_type = ARTIFACT_VERSION_CREATED_ACTOR
    attributes = ARTIFACT_VERSION_ATTRIBUTES


class ArtifactVersionUpdatedActorEvent(VersionActorEvent):
    event_type = ARTIFACT_VERSION_UPDATED_ACTOR
    attributes = ARTIFACT_VERSION_ATTRIBUTES


class ArtifactVersionDeletedActorEvent(VersionActorEvent):
    event_type = ARTIFACT_VERSION_DELETED_ACTOR
    attributes = ARTIFACT_VERSION_ATTRIBUTES


class ArtifactVersionViewedActorEvent(VersionActorEvent):
    event_type = ARTIFACT_VERSION_VIEWED_ACTOR
    attributes = ARTIFACT_VERSION_ATTRIBUTES


class ArtifactVersionTransferredActorEvent(VersionActorEvent):
    event_type = ARTIFACT_VERSION_TRANSFERRED_ACTOR
    attributes = ARTIFACT_VERSION_ATTRIBUTES


class ArtifactVersionNewStageEvent(VersionActorEvent):
    event_type = ARTIFACT_VERSION_NEW_STAGE
    attributes = ARTIFACT_VERSION_ATTRIBUTES + (
        Attribute("stage"),
        Attribute("previous_stage", is_required=False),
    )



from haupt.common.options import option_namespaces, option_subjects
from haupt.common.options.option import Option, OptionScope, OptionStores

CLEANING_INTERVALS_ACTIVITY_LOGS = "{}_{}".format(
    option_namespaces.CLEANING_INTERVALS,
    option_subjects.ACTIVITY_LOGS,
)
CLEANING_INTERVALS_NOTIFICATIONS = "{}_{}".format(
    option_namespaces.CLEANING_INTERVALS,
    option_subjects.NOTIFICATIONS,
)
CLEANING_INTERVALS_ARCHIVES = "{}_{}".format(
    option_namespaces.CLEANING_INTERVALS,
    option_subjects.ARCHIVES,
)
CLEANING_INTERVALS_DELETION = "{}_{}".format(
    option_namespaces.CLEANING_INTERVALS, option_subjects.DELETION
)


OPTIONS = {
    CLEANING_INTERVALS_ACTIVITY_LOGS,
    CLEANING_INTERVALS_NOTIFICATIONS,
    CLEANING_INTERVALS_ARCHIVES,
    CLEANING_INTERVALS_DELETION,
}


class CleaningIntervalsOption(Option):
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    typing = "int"
    store = OptionStores.SETTINGS
    options = None


class CleaningIntervalsActivityLogs(CleaningIntervalsOption):
    key = CLEANING_INTERVALS_ACTIVITY_LOGS
    default = 90
    description = "A cleaning interval for activity logs in days"


class CleaningIntervalsNotifications(CleaningIntervalsOption):
    key = CLEANING_INTERVALS_NOTIFICATIONS
    default = 30
    description = "A cleaning interval for notifications in days"


class CleaningIntervalsArchives(CleaningIntervalsOption):
    key = CLEANING_INTERVALS_ARCHIVES
    default = 30
    description = "A cleaning interval for archives in days"


class CleaningIntervalsDeletion(CleaningIntervalsOption):
    key = CLEANING_INTERVALS_DELETION
    default = 15
    description = "A cleaning interval for archives in days"



from django.db import models


class RunResources(models.Model):
    memory = models.FloatField(blank=True, null=True, default=0)
    cpu = models.FloatField(blank=True, null=True, default=0)
    gpu = models.FloatField(blank=True, null=True, default=0)
    custom = models.FloatField(blank=True, null=True, default=0)
    cost = models.FloatField(blank=True, null=True, default=0)

    class Meta:
        abstract = True



from rest_framework import fields, serializers


class ProjectOwnerMixin(serializers.Serializer):
    owner = fields.SerializerMethodField()

    def get_owner(self, obj):
        return obj.project.owner.name



from haupt.common import auditor
from haupt.common.events.registry import archive

auditor.subscribe(archive.ProjectArchivedActorEvent)
auditor.subscribe(archive.ProjectRestoredActorEvent)
auditor.subscribe(archive.RunArchivedActorEvent)
auditor.subscribe(archive.RunRestoredActorEvent)



try:
    from rest_framework.exceptions import ValidationError
except ImportError:
    raise ImportError("This module depends on django.")


NAME_BLACKLIST = {
    "user",
    "profile",
    "preset",
    "admin",
    "root",
    "abstract",
    "experiment",
    "gateway",
    "stream",
    "service",
    "project",
    "health",
    "50x",
    "404",
    "rewrite-service",
    "permission",
    "healthz",
    "api",
    "polyaxon",
    "haupt",
    "dashboard",
    "index",
    "log",
    "metric",
    "portfolio",
    "public",
    "revision",
    "version",
    "support",
    "tryout",
    "output",
    "artifact",
    "data",
    "dataset",
    "secret",
    "config_map",
    "pod",
    "deployment",
    "model",
    "management",
    "dependency",
    "dependencie",
    "repo",
    "cluster",
    "event",
    "manage",
    "login",
    "account",
    "register",
    "accept",
    "organization",
    "org",
    "team",
    "help",
    "doc",
    "logout",
    "404",
    "500",
    "static",
    "_static",
    "statusz",
    "status",
    "statuses",
    "_status",
    "out",
    "debug",
    "remote",
    "cli",
    "blog",
    "welcome",
    "feature",
    "customer",
    "integration",
    "signup",
    "pricing",
    "subscribe",
    "enterprise",
    "about",
    "job",
    "run",
    "thank",
    "guide",
    "privacy",
    "security",
    "term",
    "from",
    "sponsorship",
    "for",
    "at",
    "vs",
    "_admin",
    "support",
    "contact",
    "bookmark",
    "pipeline",
    "dag",
    "operation",
    "component",
    "action",
    "hook",
    "monitor",
    "setting",
    "hub",
    "registry",
    "registrie",
    "legal",
    "avatar",
    "self",
    "this",
    "name",
    "list",
    "detail",
    "update",
    "create",
    "delete",
    "get",
    "fetch",
    "query",
    "put",
    "patch",
    "member",
    "team",
    "new",
    "archive",
    "restore",
    "catalog",
    "setting",
    "auth",
    "activity",
    "activitie",
    "history",
    "analytic",
}
NAME_BLACKLIST |= {"{}s".format(b) for b in NAME_BLACKLIST}


def validate_blacklist_name(name: str):
    """Validates slug name against a blacklist"""
    if name is None:
        raise ValidationError("A short name must be supplied.")

    if name.lower() in NAME_BLACKLIST:
        raise ValidationError("The name is a reserved word or already taken.")



from rest_framework import status
from rest_framework.generics import CreateAPIView
from rest_framework.response import Response

from django.conf import settings

from haupt.apis.bookmarks.views import BookmarkCreateView, BookmarkDeleteView
from haupt.apis.endpoints.project import ProjectEndpoint
from haupt.apis.serializers.base.bookmarks_mixin import BookmarkedListMixinView
from haupt.apis.serializers.project_stats import ProjectStatsSerializer
from haupt.apis.serializers.projects import (
    BookmarkedProjectSerializer,
    ProjectCreateSerializer,
    ProjectDetailSerializer,
    ProjectNameSerializer,
)
from haupt.common.apis.filters import OrderingFilter, QueryFilter
from haupt.common.apis.paginator import LargeLimitOffsetPagination
from haupt.common.content_types import ContentTypes
from haupt.common.endpoints.base import (
    BaseEndpoint,
    DestroyEndpoint,
    ListEndpoint,
    PostEndpoint,
    RetrieveEndpoint,
    UpdateEndpoint,
)
from haupt.common.endpoints.mixins import StatsMixin
from haupt.common.events.registry.archive import (
    PROJECT_ARCHIVED_ACTOR,
    PROJECT_RESTORED_ACTOR,
)
from haupt.common.events.registry.bookmark import (
    PROJECT_BOOKMARKED_ACTOR,
    PROJECT_UNBOOKMARKED_ACTOR,
)
from haupt.common.events.registry.project import PROJECT_STATS_ACTOR
from haupt.db.defs import Models
from haupt.db.managers.live_state import (
    archive_project,
    delete_in_progress_project,
    restore_project,
)
from haupt.db.managers.projects import add_project_contributors
from haupt.db.managers.stats import StatsSerializer
from haupt.db.query_managers.project import ProjectQueryManager
from polyaxon._services.values import PolyaxonServices


class ProjectCreateView(BaseEndpoint, CreateAPIView):
    serializer_class = ProjectCreateSerializer
    ALLOWED_METHODS = ["POST"]


ADDITIONAL_ONLY_FIELD = ["is_public"] if settings.HAS_ORG_MANAGEMENT else []


class ProjectsListMixinView(BookmarkedListMixinView):
    queryset = Models.Project.all.only(
        "uuid",
        "name",
        "description",
        "created_at",
        "updated_at",
        "tags",
        "live_state",
        *ADDITIONAL_ONLY_FIELD,
    ).order_by("-updated_at")
    serializer_class = BookmarkedProjectSerializer

    bookmarked_model = "project"
    query_manager = ProjectQueryManager
    check_alive = ProjectQueryManager.CHECK_ALIVE
    ordering = ProjectQueryManager.FIELDS_DEFAULT_ORDERING
    ordering_fields = ProjectQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = ProjectQueryManager.FIELDS_ORDERING_PROXY

    ALLOWED_METHODS = ["GET"]


class ProjectListView(ProjectsListMixinView, BaseEndpoint, ListEndpoint):
    pass


class ProjectNameListMixin:
    queryset = Models.Project.objects.only("name").order_by("-updated_at")
    serializer_class = ProjectNameSerializer
    pagination_class = LargeLimitOffsetPagination
    filter_backends = (QueryFilter, OrderingFilter)
    query_manager = ProjectQueryManager
    check_alive = ProjectQueryManager.CHECK_ALIVE
    ordering = ProjectQueryManager.FIELDS_DEFAULT_ORDERING
    ordering_fields = ProjectQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = ProjectQueryManager.FIELDS_ORDERING_PROXY
    ALLOWED_METHODS = ["GET"]


class ProjectNameListView(ProjectNameListMixin, BaseEndpoint, ListEndpoint):
    pass


class ProjectDetailView(
    ProjectEndpoint, RetrieveEndpoint, UpdateEndpoint, DestroyEndpoint
):
    queryset = (
        Models.Project.all.select_related("owner").prefetch_related("contributors")
        if settings.HAS_ORG_MANAGEMENT
        else Models.Project.all
    )
    serializer_class = ProjectDetailSerializer
    ALLOWED_METHODS = ["GET", "PUT", "PATCH", "DELETE"]
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_INSTANCE = True

    def perform_update(self, serializer):
        instance = serializer.save()
        add_project_contributors(instance, users=[self.request.user])

    def perform_destroy(self, instance):
        delete_in_progress_project(instance)


class ProjectArchiveView(ProjectEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": PROJECT_ARCHIVED_ACTOR}
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        obj = self.get_object()
        add_project_contributors(obj, users=[self.request.user])
        archive_project(obj)
        self.audit(request, *args, **kwargs)
        return Response(status=status.HTTP_200_OK, data={})


class ProjectRestoreView(ProjectEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": PROJECT_RESTORED_ACTOR}
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        obj = self.get_object()
        add_project_contributors(obj, users=[self.request.user])
        self.audit(request, *args, **kwargs)
        restore_project(obj)
        return Response(status=status.HTTP_200_OK, data={})


class ProjectBookmarkCreateView(ProjectEndpoint, BookmarkCreateView):
    content_type = ContentTypes.PROJECT.value
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": PROJECT_BOOKMARKED_ACTOR}
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_INSTANCE = True


class ProjectBookmarkDeleteView(ProjectEndpoint, BookmarkDeleteView):
    content_type = ContentTypes.PROJECT.value
    ALLOWED_METHODS = ["DELETE"]
    AUDITOR_EVENT_TYPES = {"DELETE": PROJECT_UNBOOKMARKED_ACTOR}
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_INSTANCE = True


class ProjectStatsView(ProjectEndpoint, RetrieveEndpoint, StatsMixin):
    queryset = (
        Models.Project.all.select_related("owner")
        if settings.HAS_ORG_MANAGEMENT
        else Models.Project.all
    )
    AUDITOR_EVENT_TYPES = {
        "GET": PROJECT_STATS_ACTOR,
    }
    ALLOWED_SERVICES = [PolyaxonServices.UI]
    CHECK_SERVICE = True
    ALLOWED_METHODS = ["GET"]
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_INSTANCE = True
    bookmarked_model = "run"

    def get_queryset(self):
        mode = self.validate_stats_mode()
        queryset = super().get_queryset()
        if mode == "stats":
            return queryset.select_related("latest_stats")
        return queryset

    def get_serializer(self, *args, **kwargs):
        mode = self.validate_stats_mode()
        if mode == "stats":
            return ProjectStatsSerializer(self.project.latest_stats)
        queryset = Models.Run.objects.filter(project_id=self.project.id)
        queryset = StatsSerializer.filter_queryset(
            queryset=queryset,
            request=self.request,
            view=self,
        )
        return StatsSerializer(
            queryset=queryset,
            kind=self.request.query_params.get("kind"),
            aggregate=self.request.query_params.get("aggregate"),
            groupby=self.request.query_params.get("groupby"),
            trunc=self.request.query_params.get("trunc"),
        )



from django.db import models

from haupt.db.abstracts.stats import StatsModel


class ProjectStats(StatsModel):
    project = models.ForeignKey(
        "db.Project", on_delete=models.CASCADE, related_name="stats"
    )

    class Meta(StatsModel.Meta):
        db_table = "db_projectstats"
